
#ifndef BFM_CG_H
#define BFM_CG_H

/*  2007
 *  Calls PAB's assembler routines to give an implementation
 *  of the dwf dslash. Aim is to give very high performance
 *  in a reasonably portable/retargettable library.
 */
#include "bfm.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#ifdef LIST_ENGINE
//include <spi/include/l1p/pprefetch.h>
#include <pprefetch.h>
#endif

template <class Float>
int bfmbase<Float>::mprecFlopsPerSite ( void ) 
{
  int fourdwilson = 1320;
  int fivedhop    = 24;

  int flops;

  if ( precon_5d && (Ls > 1) ) {
    // A + B d d
    flops =  2*(fourdwilson + fivedhop) + 72;
  } else if ( Ls > 1 ) {
    flops = 2*(fourdwilson + 24*5);
  } else {
    // A+ B d d 
    flops =  2*fourdwilson + 72;
  }
  return flops;
}

template <class Float>
int bfmbase<Float>::cbSites( void ) 
{
      int lx = node_latt[0];
      int ly = node_latt[1];
      int lz = node_latt[2];
      int lt = node_latt[3];

      int sites = (lx*ly*lz*lt*Ls)/2;

      return sites;
}

template <class Float>
int bfmbase<Float>::mprecFlops( void ) 
{
  return mprecFlopsPerSite() * cbSites();
}
template <class Float>
int bfmbase<Float>::axpyFlops ( void ) 
{
  return 2*24*cbSites();
}
template <class Float>
int bfmbase<Float>::axpyNormFlops ( void ) 
{
  return 4*24*cbSites();
}
template <class Float>
int bfmbase<Float>::axpyBytes ( void ) 
{
  return 8*24*3*cbSites();
}


  /*
   * Red black Schur decomposition
   *
   *  M = (Mee Meo) =  (1             0 )   (Mee   0               )  (1 Mee^{-1} Meo)
   *      (Moe Moo)    (Moe Mee^-1    1 )   (0   Moo-Moe Mee^-1 Meo)  (0   1         )
   *                =         L                     D                     U
   *
   * L^-1 = (1              0 )
   *        (-MoeMee^{-1}   1 )   
   * L^{dag} = ( 1       Mee^{-dag} Moe^{dag} )
   *           ( 0       1                    )
   * L^{-d}  = ( 1      -Mee^{-dag} Moe^{dag} )
   *           ( 0       1                    )
   *
   * U^-1 = (1   -Mee^{-1} Meo)
   *        (0    1           )
   * U^{dag} = ( 1                 0)
   *           (Meo^dag Mee^{-dag} 1)
   * U^{-dag} = (  1                 0)
   *            (-Meo^dag Mee^{-dag} 1)
   ***********************************************************
   */

  /*
   *
   * Routines:
   *
   *  bfm::CGNE_M 
   *  bfm::CGNE (historical name for CGNE_M)
   ***********************
   *     M psi = eta
   ***********************
   *Odd
   * i)   (D_oo)^{\dag} D_oo psi_o = (D_oo)^\dag L^{-1}  eta_o
   *                        eta_o' = D_oo (eta_o - Moe Mee^{-1} eta_e)
   *Even
   * ii)  Mee psi_e + Meo psi_o = src_e
   *
   *   => sol_e = M_ee^-1 * ( src_e - Meo sol_o )...
   *
   */

template <class Float>
int bfmbase<Float>::CGNE(Fermion_t solution[2], Fermion_t source[2], multi1d<bfm_fermion> &evecs, multi1d<double> &evals)
{
  return CGNE_M(solution,source, evecs, evals);
}
template <class Float>
int bfmbase<Float>::CGNE(Fermion_t solution[2], Fermion_t source[2])
{
  return CGNE_M(solution,source);
}
template <class Float>
int bfmbase<Float>::CGNE_M(Fermion_t solution[2], Fermion_t source[2])
{
  multi1d<bfm_fermion> evecs(0);
  multi1d<double> evals(0);
  return CGNE_M(solution,source, evecs,evals);
}
template <class Float>
int bfmbase<Float>::CGNE_M(Fermion_t solution[2], Fermion_t source[2], multi1d<bfm_fermion> &evecs, multi1d<double> &evals)
{
  int me = thread_barrier();
  Fermion_t src = threadedAllocFermion(); 
  Fermion_t hsrc = threadedAllocFermion(); 
  Fermion_t tmp = threadedAllocFermion(); 
  Fermion_t Mtmp= threadedAllocFermion(); 
  Fermion_t lsol= threadedAllocFermion(); 

  if (isBoss() && !me ) printf("bfmbase::CGNE: Solver type is %d\n",solver);

  double f;
  f = norm(source[0]);
  f+= norm(source[1]);

  if (isBoss() && !me ) printf("bfmbase::CGNE: Source norm is %le\n",f);
  f = norm(solution[0]);
  f+= norm(solution[1]);
  if (isBoss() && !me ) printf("bfmbase::CGNE: Guess norm is %le\n",f);

  // src_o = Mdag * (source_o - Moe MeeInv source_e)
  //
  // When using the CGdiagonalMee, we left multiply system of equations by 
  // diag(MeeInv,MooInv), and must multiply by an extra MooInv
  //

  if ( CGdiagonalMee ) { 
    MooeeInv(source[Even],tmp,DaggerNo);
    Meo     (tmp,src,Odd,DaggerNo);
    axpy    (src,src,source[Odd],-1.0);
    MooeeInv(src,tmp,DaggerNo);
    Mprec(tmp,src,Mtmp,DaggerYes);  
  } else { 
    MooeeInv(source[Even],tmp,DaggerNo);
    Meo     (tmp,src,Odd,DaggerNo);
    axpy    (tmp,src,source[Odd],-1.0);
    Mprec(tmp,src,Mtmp,DaggerYes);  
  }
  /*
  MooeeInv(source[Even],tmp,DaggerNo);
  f = norm(tmp);
  if (isBoss() && !me ) printf("bfmbase::CGNE: prec tmp norm (MooeeInv) is %le\n",f);
  Meo     (tmp,src,Odd,DaggerNo);
  f = norm(src);
  if (isBoss() && !me ) printf("bfmbase::CGNE: prec src norm (Meo) is %le\n",f);
  axpy    (tmp,src,source[Odd],-1.0);
  f = norm(tmp);
  if (isBoss() && !me ) printf("bfmbase::CGNE: prec tmp norm (axpy) is %le\n",f);
  Mprec(tmp,src,Mtmp,DaggerYes);  
  f = norm(src);
  if (isBoss() && !me ) printf("bfmbase::CGNE: prec src norm (MprecDag) is %le\n",f);
  */  
  int Nev = evals.size();
  if(Nev > 0){

    if ( isBoss() && !me ) printf("bfmbase::CGNE: deflating with %d evecs\n",Nev);

    axpy    (hsrc,src,src,0.0);
    axpby   (lsol,src,src,0.0,0.0);
  
    for(int n = 0; n<Nev; n++){
      complex<double> cn = inner (evecs[n][1], src);
      ///deflate src
      axpy    (Mtmp,evecs[n][1],evecs[n][1],0.0);
      scale   (Mtmp,real(cn),imag(cn) );
      axpy    (hsrc,Mtmp,hsrc,-1.0);
      ///low solution
      axpy    (lsol, Mtmp, lsol, 1.0/evals[n]);
    }
  
    // Do we want to be able to control whether
    // i)   guess becomes hsrc.
    // ii)  guess becomes zero
    // iii) guess becomes orthogonal subspace projection of guess (costs a bit more)
    //      but means a perfect guess finishes and is useful to restarted CG.
    //      Importance is reduced once I implement defect correction single/double CG.

    axpy(solution[Odd], hsrc, hsrc, 0.0);  

  }

  int iter = CGNE_prec_MdagM(solution[Odd],src);
  
  if ( Nev > 0 ) { 
    axpy(solution[Odd], lsol, solution[Odd], 1.0);  
  }
 
  threadedFreeFermion(Mtmp);
  threadedFreeFermion(lsol);
  threadedFreeFermion(hsrc);

  // sol_e = M_ee^-1 * ( src_e - Meo sol_o )...
  Meo(solution[Odd],tmp,Even,DaggerNo);
  axpy(src,tmp,source[Even],-1.0);
  MooeeInv(src,solution[Even],DaggerNo);
  
  f =norm(solution[0]);
  f+=norm(solution[1]);
  if (isBoss() && !me ) printf("bfmbase::CGNE: unprec sol norm is %le\n",f);
  
  threadedFreeFermion(tmp);
  threadedFreeFermion(src);

  return iter;
}

  /*
   ***********************
   * Routine:
   *
   *  bfm::CGNE_Mdag
   *
   ***********************
   *     Mdag psi = U^dag, D^dag L^{dag} = eta
   ***********************
   *
   *Odd
   * i)  D_oo^dagD_oo phi_o = U^{-dag} eta_o
   *
   *     eta_o' = eta_o - Moe^{dag}  Mee^{-dag} eta_e
   *
   *     phi_o = D_oo^{-1} Doo^{-dag} eta_o'
   *
   *     psi_o = D_oo phi_o 
   *
   *Even
   *ii) 
   *     Mee^{d} psi_e + Meo^{d} psi_o = src_e
   * =>  psi_e = Mee^{-d} ( src_e - Meo^{d} psi_o)
   *
   */

template <class Float>
int bfmbase<Float>::CGNE_Mdag(Fermion_t solution[2], Fermion_t source[2])
{
  Fermion_t src = threadedAllocFermion(); 
  Fermion_t tmp = threadedAllocFermion(); 

  // src_o = source_o - Moe^d MeeInv^d source_e
  MooeeInv(source[Even],src,DaggerYes);
  Meo     (src,tmp,Odd,DaggerYes);
  axpy    (src,tmp,source[Odd],-1.0);

  //  phi_o = D_oo^{-1} Doo^{-dag} eta_o'
  int iter = CGNE_prec_MdagM(tmp,src);

  //     psi_o = D_oo phi_o 
  Mprec(tmp,solution[Odd],src,DaggerNo);  
  
  // psi_e = Mee^{-d} ( src_e - Meo^{d} psi_o)
  Meo(solution[Odd],tmp,Even,DaggerYes);
  axpy(src,tmp,source[Even],-1.0);
  MooeeInv(src,solution[Even],DaggerYes);

  threadedFreeFermion(tmp);
  threadedFreeFermion(src);

  return iter;
}

  /*
   *
   ***********************
   * Routine:
   *  bfm::CGNE_prec_MdagM
   ***********************
   *     MdagM psi = eta
   ***********************
   *   Simply solves
   *   D_oo^dagD_oo phi_o = L^{-1} eta_o
   *
   */


template <class Float>
int bfmbase<Float>::CGNE_prec_MdagM(Fermion_t psi, Fermion_t src)
{
  int iter;
  int me= thread_barrier();

  if ( reproduce ) { 

    uint64_t c1,c2;

    Fermion_t tmp_r = threadedAllocFermion(); 
    axpy    (tmp_r,psi,psi,0.0);

    c1=checksum(tmp_r);
    c2=checksum(psi);

    repro_control(ReproRecord);
    iter = CGNE_prec(tmp_r,src);
    if ( isBoss() && !me ) printf("bfmbase:: Checking reproducibility\n");
    repro_control(ReproCheck);
    iter = CGNE_prec(psi,src);
    repro_control(ReproIdle);

    c1=checksum(tmp_r);
    c2=checksum(psi);
    if ( c1!=c2  ) {
       printf("bfmbase:: result vector checksum miscompare in reproducing CG %16.16lx %16.16lx\n",c1,c2);
       exit(-1);
    }
    if ( isBoss() && !me ) printf("bfmbase:: reproducing cg passed -- checksums are %16.16lx %16.16lx\n",c1,c2);

    threadedFreeFermion(tmp_r);

  } else {
    double guess = norm (psi);

    int me = this->thread_barrier();

    if ( verbose && isBoss() && !me ) {
      printf("Not reproducing -- guess is %le\n",guess);
    }
    iter = CGNE_prec(psi,src);
    repro_control(ReproIdle);

  }
  return iter;
}

template <class Float>
int bfmbase<Float>::CGNE_prec(Fermion_t psi, Fermion_t src)
{

  double f;
  double cp,c,a,d,b;
  int me = thread_barrier();

  if ( this->isBoss() && (!me) ) { 
    this->InverterEnter();
  }

  Fermion_t p   = threadedAllocFermion(mem_fast); 
  Fermion_t tmp = threadedAllocFermion(mem_fast); 
  Fermion_t mp  = threadedAllocFermion(mem_fast); 
  Fermion_t mmp = threadedAllocFermion(mem_fast); 
  Fermion_t r   = threadedAllocFermion(mem_fast); 

  //Initial residual computation & set up
  double guess = norm(psi);
  d=Mprec(psi,mp,tmp,DaggerNo);
  b=Mprec(mp,mmp,tmp,DaggerYes);

  axpy (r, mmp, src,-1.0);
  axpy (p, mmp, src,-1.0);

  a =norm(p);
  cp= norm(r);

  Float ssq =  norm(src);
  if ( verbose && isBoss() && !me ) {
    printf("bfmbase::CGNE_prec gues %le \n",guess);
    printf("bfmbase::CGNE_prec src  %le \n",ssq);
    printf("bfmbase::CGNE_prec  mp  %le \n",d);
    printf("bfmbase::CGNE_prec  mmp %le \n",b);
    printf("bfmbase::CGNE_prec   r  %le \n",cp);
    printf("bfmbase::CGNE_prec   p  %le \n",a);
  }
  Float rsq =  residual* residual*ssq;

  //Check if guess is really REALLY good :)
  if ( cp <= rsq ) {
    if ( verbose && isBoss() && !me ) {
      printf("bfmbase::CGNE_prec k=0 converged - suspiciously nice guess %le %le\n",cp,rsq);
    }
    threadedFreeFermion(tmp);
    threadedFreeFermion(p);
    threadedFreeFermion(mp);
    threadedFreeFermion(mmp);
    threadedFreeFermion(r);
    if ( this->isBoss() && (!me) ) { 
      this->InverterExit();
    }
    return 0;
  }
  if ( verbose && isBoss() && !me ) 
    printf("bfmbase::CGNE_prec k=0 residual %le rsq %le\n",cp,rsq);

  if ( isBoss() && !me ) {
    if ( this->watchfile ) {
      printf("bfmbase::CGNE_prec watching file \"%s\"\n",this->watchfile);
    }    
  }
  struct timeval start,stop;
  gettimeofday(&start,NULL);


  for (int k=1;k<=max_iter;k++){

    this->iter=k;
    uint64_t t_iter_1=GetTimeBase();

    c=cp;

    uint64_t t_mprec_1=GetTimeBase();
    d = Mprec(p,mp,tmp,0,1);
    uint64_t t_mprec_2=GetTimeBase();
    a = c/d;

    //NB: No longer need mp, overwrite
    //    With optimised version need to take
    //            mp-> Dmp
    //        Dmp,mp-> amp+DDmp -> mp inplace.
    // Implement "Mprec_inplace(mp,mmp,1); 
    uint64_t t_mprec_3=GetTimeBase();
    double qq=Mprec(mp,mmp,tmp,1); 
    uint64_t t_mprec_4=GetTimeBase();
    double b_pred = a*(a*qq-d)/c;

#if 0
    // Can fuse these together. 
    // Further, producing "p" in a cache-blocked fashion could allow no reload
    // of p from main memory (~8 timeslices at a time?)

    //Fused axpy-norm required.    
    //As know "a" in advance could save bandwidth by
    //not writing out mmp, but doing linear combination.
    //in Mprec.
    uint64_t tr1=GetTimeBase();
    cp = axpy_norm(r,mmp,r,-a);
    b = cp/c;
    uint64_t tr2=GetTimeBase();

    uint64_t tpsi1=GetTimeBase();
    axpy(psi,p,psi,a);
    uint64_t tpsi2=GetTimeBase();
    // New (conjugate/M-orthogonal) search direction
    uint64_t tp1=GetTimeBase();
    axpy(p,p,r,b);
    uint64_t tp2=GetTimeBase();
#else
    uint64_t tpsi1=GetTimeBase();
    uint64_t tp1=GetTimeBase();
    uint64_t tp2=tp1+1;
    uint64_t tr1=tp1;
    uint64_t tr2=tp2;
    b  = b_pred;
    cp = cg_psi_p(psi,p,r,mmp,a,b);
    uint64_t tpsi2=GetTimeBase();
#endif
    uint64_t t_iter_2=GetTimeBase();

    // Inefficient hack while I test something
    if(0){
      if ( this->isBoss() && !me ) {
	printf("k=%d alpha = %le\n",k,a);
	printf("k=%d predict beta=%le\n",k,b_pred);
	printf("k=%d beta = %le\n",k,b);
      }
    }
    
    // verbose nonsense
    if ( (this->iter==this->time_report_iter) && isBoss() && (!me)  && verbose) {
      
      int lx = node_latt[0];
      int ly = node_latt[1];
      int lz = node_latt[2];
      int lt = node_latt[3];

      int cb4dsites = (lx*ly*lz*lt)/2;
      printf("fermionCacheFootprint: %ld \n",7*axpyBytes()/3);
      printf("gauge  CacheFootprint: %ld \n",2*18*8*cb4dsites*2);
      printf("fermionVecBytes      : %ld \n",axpyBytes()/3);
      //printf("axpyBytes            : %ld \n",axpyBytes());
      //printf("axpy      (soln)     : %ld cyc %le MB/s\n",(tpsi2-tpsi1),(double)axpyBytes()*1600./(tpsi2-tpsi1));
      //printf("axpy_norm (residual) : %ld cyc %le MB/s\n",(tr2-tr1),(double)axpyBytes()*1600./(tr2-tr1));
      //printf("axpy      (search)   : %ld cyc %le MB/s\n",(tp2-tp1),(double)axpyBytes()*1600./(tp2-tp1));
      printf("Iter time            : %ld cyc\n",t_iter_2-t_iter_1);
      printf("linalg time          : %ld cyc\n",t_iter_2-t_iter_1-(t_mprec_2-t_mprec_1)-(t_mprec_4-t_mprec_3));
      printf("Mprec time           : %ld cyc\n",t_mprec_2-t_mprec_1);
      printf("Mprec time           : %ld cyc\n",t_mprec_4-t_mprec_3);
      fflush(stdout);
    }


    if ( ((k%100 == 0) && (verbose!=0)) || (verbose > 10)){
      if ( isBoss() && !me ) {
	printf("bfmbase::CGNE_prec: k=%d r^2=%le %le %lx\n",k,cp,sqrt(cp/ssq),this);
      }
    }

    //if ( k%100 ) {
      //      if ( this->CheckStopWatchFile() ) {
      //	printf("bfmbase::CGNE_prec: terminating due to stop request %s\n",this->watchfile);
      //abort();
      //      }
    //}

    // Stopping condition
    if ( cp <= rsq ) { 

      gettimeofday(&stop,NULL);
      struct timeval diff;
      timersub(&stop,&start,&diff);

      if ( isBoss() && !me ) printf("bfmbase::CGNE_prec converged in %d iterations\n",k);
      if ( isBoss() && !me ) printf("bfmbase::CGNE_prec converged in %d.%6.6d s\n",diff.tv_sec,diff.tv_usec);


      double flops = mprecFlops()*2.0 + 2.0*axpyNormFlops() + axpyFlops()*2.0;
             flops = flops * k;

      double t = diff.tv_sec*1.0E6 + diff.tv_usec;
      if ( isBoss()&& !me ) 
       printf("bfmbase::CGNE_prec: %d mprec flops/site\n",mprecFlopsPerSite());
      if ( isBoss()&& !me ) printf("bfmbase::CGNE_prec: %le flops\n",flops);
      if ( isBoss()&& !me ) printf("bfmbase::CGNE_prec: %le mflops per node\n",flops/t);

      Mprec(psi,mp,tmp,0);
      Mprec(mp,mmp,tmp,1); 
      axpy(tmp,src,mmp,-1.0);
      double true_residual = sqrt(norm(tmp)/norm(src));
      if ( isBoss() && !me ) 
	printf("bfmbase::CGNE_prec: true residual is %le \n",true_residual);

      threadedFreeFermion(tmp);
      threadedFreeFermion(p);
      threadedFreeFermion(mp);
      threadedFreeFermion(mmp);
      threadedFreeFermion(r);
#ifdef LIST_ENGINE
      if ( this->list_engine) L1P_PatternUnconfigure();
#endif
      if ( this->isBoss() && (!me) ) { 
	this->InverterExit();
      }
      return k;
    }

  }
  if ( isBoss() && !me ) printf("bfmbase::CGNE_prec: CG not converged \n");
  threadedFreeFermion(tmp);
  threadedFreeFermion(p);
  threadedFreeFermion(mp);
  threadedFreeFermion(mmp);
  threadedFreeFermion(r);
#ifdef LIST_ENGINE
  if (this->list_engine ) L1P_PatternUnconfigure();
#endif
  if ( this->isBoss() && (!me) ) { 
    this->InverterExit();
  }

  return -1;

}


/**********************************************************
*                                                         *
*           PV taku trick. Can probably do this better    *
*                                                         *
***********************************************************/

template <class Float>
int bfmbase<Float>::CGNE_PV_M(Fermion_t solution[2], Fermion_t source[2],int n, int L5)
{
  Fermion_t src = threadedAllocFermion(); 
  Fermion_t tmp = threadedAllocFermion(); 
  Fermion_t Mtmp= threadedAllocFermion(); 

  double f;
  f = norm(source[0]);
  f+= norm(source[1]);

  int me = thread_barrier();
  //if (isBoss() && !me ) printf("bfmbase::CGNE: Source norm is %le\n",f);

  f = norm(solution[0]);
  f+= norm(solution[1]);

  //if (isBoss() && !me ) printf("bfmbase::CGNE: Guess norm is %le\n",f);

  // src_o = Mdag * (source_o - Moe MeeInv source_e)
  MooeeInv5dprec_TW(source[Even],tmp,n,L5,DaggerNo);
  f = norm(tmp);

  //if (isBoss()  && !me) printf("bfmbase::CGNE: tmp norm is %le\n",f);
  Meo     (tmp,src,Odd,DaggerNo);
  f = norm(src);

  //if (isBoss() && !me) printf("bfmbase::CGNE: dsl norm is %le\n",f*4);
  axpy    (tmp,src,source[Odd],-1.0);
  
  MprecTW(tmp,src,Mtmp,DaggerYes,n,L5,0);  
  threadedFreeFermion(Mtmp);
  
  int iter = CGNE_PV_prec_M(solution[Odd],src,n,L5);
  axpby(solution[Even],tmp,tmp,0.0,0.0);
  
  // sol_e = M_ee^-1 * ( src_e - Meo sol_o )...
  Meo(solution[Odd],tmp,Even,DaggerNo);
  axpy(src,tmp,source[Even],-1.0);
  MooeeInv5dprec_TW(src,solution[Even],n,L5,DaggerNo);


  threadedFreeFermion(tmp);
  threadedFreeFermion(src);

  return iter;
}

template <class Float>
int bfmbase<Float>::CGNE_PV_Mdag(Fermion_t solution[2], Fermion_t source[2],int n, int L5)
{
  Fermion_t src = threadedAllocFermion(); 
  Fermion_t tmp = threadedAllocFermion(); 

  // src_o = source_o - Moe^d MeeInv^d source_e
  MooeeInv5dprec_TW(source[Even],src,n,L5,DaggerYes);
  Meo     (src,tmp,Odd,DaggerYes);
  
  axpy    (src,tmp,source[Odd],-1.0);

  //  phi_o = D_oo^{-1} Doo^{-dag} eta_o'
  int iter = CGNE_PV_prec_M(tmp,src, n, L5);

  //     psi_o = D_oo phi_o 
  MprecTW(tmp,solution[Odd],src,DaggerNo,n,L5,0);  
  
  // psi_e = Mee^{-d} ( src_e - Meo^{d} psi_o)
  Meo(solution[Odd],tmp,Even,DaggerYes);
  
  axpy(src,tmp,source[Even],-1.0);
  MooeeInv5dprec_TW(src,solution[Even],n,L5,DaggerYes);

  threadedFreeFermion(tmp);
  threadedFreeFermion(src);

  return iter;
}

template <class Float>
int bfmbase<Float>::CGNE_PV_prec_M(Fermion_t psi, Fermion_t src,int n, int L5)
{
  int iter;
  int me= thread_barrier();

  if ( reproduce ) { 

    uint64_t c1,c2;

    Fermion_t tmp_r = threadedAllocFermion(); 
    axpy    (tmp_r,psi,psi,0.0);


    c1=checksum(tmp_r);
    c2=checksum(psi);

    repro_control(ReproRecord);
    iter = CGNE_PV_prec(tmp_r,src,n,L5);
    if ( isBoss() && !me ) printf("bfmbase:: Checking reproducibility\n");
    repro_control(ReproCheck);
    iter = CGNE_PV_prec(psi,src,n,L5);
    repro_control(ReproIdle);

    c1=checksum(tmp_r);
    c2=checksum(psi);
    if ( c1!=c2  ) {
       printf("bfmbase:: result vector checksum miscompare in reproducing CG %16.16lx %16.16lx\n",c1,c2);
       exit(-1);
    }
    if ( isBoss() && !me ) printf("bfmbase:: reproducing cg passed -- checksums are %16.16lx %16.16lx\n",c1,c2);

    threadedFreeFermion(tmp_r);

  } else {
    iter = CGNE_PV_prec(psi,src,n,L5);
    repro_control(ReproIdle);

  }
  return iter;
}


template <class Float>
int bfmbase<Float>::CGNE_PV_prec(Fermion_t psi, Fermion_t src,int n, int L5)
{

  double f;
  double cp,c,a,d,b;
  int me = thread_barrier();

  Fermion_t p   = threadedAllocFermion(mem_fast); 
  Fermion_t tmp = threadedAllocFermion(mem_fast); 
  Fermion_t mp  = threadedAllocFermion(mem_fast); 
  Fermion_t mmp = threadedAllocFermion(mem_fast); 
  Fermion_t r   = threadedAllocFermion(mem_fast); 

  //Initial residual computation & set up
  MprecTW(psi,mp,tmp,DaggerNo,n,L5,0);
  MprecTW(mp,mmp,tmp,DaggerYes,n,L5,0);



  axpy (r, mmp, src,-1.0);
  axpy (p, mmp, src,-1.0);

  cp= norm(r);
  Float ssq =  norm(src);
  Float rsq =  residual* residual*ssq;
  
  //Check if guess is really REALLY good :)
  if ( cp <= rsq ) {
    if ( isBoss() && !me )  printf("bfmbase::CGNE_prec k=0 converged - nice guess %le %le\n",cp,rsq);
    
    threadedFreeFermion(tmp);
    threadedFreeFermion(p);
    threadedFreeFermion(mp);
    threadedFreeFermion(mmp);
    threadedFreeFermion(r);
    return 0;
  }
  //if ( isBoss() && !me ) printf("bfmbase::CGNE_prec k=0 residual %le rsq %le\n",cp,rsq);

  struct timeval start,stop;
  gettimeofday(&start,NULL);

#ifdef LIST_ENGINE
  if (this->list_engine) {
    int rc = 0;
    rc = L1P_PatternConfigure(4096);
    if(rc <0 ) {
      printf("bfmbase: Bad list configure\n");
      exit(-1);
    }
    uint64_t *pf_usr = (uint64_t *)L1P_CFG_PF_USR;
    *pf_usr |=            L1P_CFG_PF_USR_pf_stream_est_on_dcbt 
      | L1P_CFG_PF_USR_pf_stream_optimistic  
      | L1P_CFG_PF_USR_pf_stream_prefetch_enable
      | L1P_CFG_PF_USR_pf_stream_establish_enable;
    thread_barrier();
    if ( isBoss() && !me ) {
      printf("bfmbase:: LIST Patterns configured\n"); fflush(stdout);
    }
    thread_barrier();
  }
#endif

  for (int k=1;k<=max_iter;k++){
    c=cp;

    // Mp
#ifdef LIST_ENGINE
    if (this->list_engine)  L1P_PatternStart(1);
    
    //thread_barrier();
    //if ( isBoss() && !me ) {
     // L1P_PatternPause();
     // printf("started list engine\n");
     // L1P_PatternResume();
    //}
    //thread_barrier();
    
#endif
    d = MprecTW(p,mp,tmp,0,n,L5,1);

    // Update solution vector
    a = c/d;
    axpy(psi,p,psi,a);
    
    // MdagMp
    MprecTW(mp,mmp,tmp,1,n,L5,0); 
#ifdef LIST_ENGINE
    
    //L1P_PatternPause();
    //thread_barrier();
    //if ( isBoss() && !me ) printf("paused list engine\n");
    //thread_barrier();
    //if ( isBoss() && !me ) printf("stopping list engine\n");
    //thread_barrier();
    
    if (this->list_engine)  L1P_PatternStop();
    //    thread_barrier();
#endif

    //Fused axpy-norm required.    
    //As know "a" in advance could save bandwidth by
    //not writing out mmp, but doing linear combination.
    //in Mprec.
    cp = axpy_norm(r,mmp,r,-a);
   // if ( ((k%20 == 0) && (verbose!=0)) || (verbose > 2)){
      //if ( isBoss() && !me ) {
   //	printf("bfmbase::CGNE_prec: k=%d r^2=%le %le %lx\n",k,cp,sqrt(cp/ssq),this);
   //     }
   //  }
   // printf("bfmbase::CGNE_prec: k=%d r^2=%le %le %lx\n",k,cp,sqrt(cp/ssq),this);
    
    // Stopping condition
    if ( cp <= rsq ) { 

      gettimeofday(&stop,NULL);
      struct timeval diff;
      timersub(&stop,&start,&diff);

      if ( isBoss() && !me ) printf("bfmbase::CGNE_prec converged in %d iterations\n",k);
      //if ( isBoss() && !me ) printf("bfmbase::CGNE_prec converged in %d.%6.6d s\n",diff.tv_sec,diff.tv_usec);

      int lx = node_latt[0];
      int ly = node_latt[1];
      int lz = node_latt[2];
      int lt = node_latt[3];
      int sites = (lx*ly*lz*lt*Ls)/2;

      //double flops = 1.0*sites*k*(  2.0*mprecFlopsPerSite() // Two matrix multiples
	//			    + 48*3                  // Three AXPY's
	//			    + 48*2  );              // Two   Norms (one in Mprec)

      //double t = diff.tv_sec*1.0E6 + diff.tv_usec;
      //if ( isBoss()&& !me ) printf("bfmbase::CGNE_prec: %d mprec flops/site\n",mprecFlopsPerSite());
      //if ( isBoss()&& !me ) printf("bfmbase::CGNE_prec: %le flops\n",flops);
      //if ( isBoss()&& !me ) printf("bfmbase::CGNE_prec: %le mflops\n",flops/t);

      MprecTW(psi,mp,tmp,0,n,L5,0);
      MprecTW(mp,mmp,tmp,1,n,L5,0); 
      axpy(tmp,src,mmp,-1.0);
      double true_residual = sqrt(norm(tmp)/norm(src));
      if ( isBoss() && !me ) printf("bfmbase::CGNE_prec: n = %d true residual is %le \n",n,true_residual);

      threadedFreeFermion(tmp);
      threadedFreeFermion(p);
      threadedFreeFermion(mp);
      threadedFreeFermion(mmp);
      threadedFreeFermion(r);
#ifdef LIST_ENGINE
      if ( this->list_engine) L1P_PatternUnconfigure();
#endif
      return k;
    }
    // New (conjugate/M-orthogonal) search direction
    b = cp/c;
    axpy(p,p,r,b);

  }
  if ( isBoss() && !me ) printf("bfmbase::CGNE_prec: CG not converged \n");
  threadedFreeFermion(tmp);
  threadedFreeFermion(p);
  threadedFreeFermion(mp);
  threadedFreeFermion(mmp);
  threadedFreeFermion(r);
#ifdef LIST_ENGINE
  if (this->list_engine ) L1P_PatternUnconfigure();
#endif
  return -1;
}


/*************************************************************************
**************************************************************************
**************************************************************************/


#define MULTI_CG_SHIFTED_CPS
#ifdef MULTI_CG_SHIFTED_CPS
#include "bfm_cg_multi.h"
#else 
template <class Float>
int bfmbase<Float>::CGNE_prec_MdagM_multi_shift(Fermion_t psi[], 
						Fermion_t src,
						double    mass[],
						double    alpha[],
						int       nshift,
						double mresidual[],
						int single)
{
  int me = thread_barrier();
 
  
  // Per shift fields
  Fermion_t ps [nshift]; // search directions 
  double    bs [nshift];
  double    rsq[nshift];
  double    z[nshift][2];
  int       converged[nshift];

  const int       primary =0;

  //Primary shift fields CG iteration
  double a,b,c,d;
  double cp,bp; //prev
  Fermion_t p ;
  Fermion_t r ;

  // Matrix mult fields
  Fermion_t tmp = threadedAllocFermion(mem_fast); 
  Fermion_t mp  = threadedAllocFermion(mem_fast); 
  Fermion_t mmp = threadedAllocFermion(mem_fast); 

  // Check lightest mass
  for(int s=0;s<nshift;s++){
    if ( mass[s] < mass[primary] ) {
      printf("First shift not lightest - oops\n");
      exit(-1);
    }
  }

  //Allocate per shift fields
  r = threadedAllocFermion(mem_slow);
  p = threadedAllocFermion(mem_slow);
  for(int i=0;i<nshift;i++){
    ps[i] = threadedAllocFermion(mem_slow);
    converged[i]=0;
  }

  // Wire guess to zero
  // Residuals "r" are src
  // First search direction "p" is also src
  cp = norm(src);
  for(int s=0;s<nshift;s++){
    rsq[s] = cp * mresidual[s] * mresidual[s];
    if ( isBoss() && (!me) && verbose ) printf("bfmbase::CGNE_prec_multi: shift %d target resid %le %le %le\n",s,rsq[s],cp,mresidual[s]);
    copy(ps[s],src);
  }
  // r and p for primary
  copy(r,src);
  copy(p,src);
  
  //MdagM+m[0]
  d= Mprec(p,mp,tmp,DaggerNo,1); 
     Mprec(mp,mmp,tmp,DaggerYes);

  b = -cp /d;


  // Set up the various shift variables
  int       iz=0;
  for(int s=0;s<nshift;s++){
    z[s][1-iz] = 1.0;
    z[s][iz]   = 1.0/( 1.0 - b*mass[s]);
    bs[s]      = b*z[s][iz]; // Sign relative to Mike - FIXME
  }
  
  // r += b[0] A.p[0]
  // c= norm(r)
  c=axpy_norm(r,mmp,r,b);

  if ( isBoss() && !me ) printf("bfmbase::CGNE_prec_multi: k=0 residual %le \n",c);

  // Linear algebra overhead can be minimised if summing results
  // psi-= b[0]p[0] = -b[0]src

  //  if ( single ) {
  //    double coeff=0;
  //    for(int s=0;s<nshift;s++) coeff+=-bs[s]*alpha[s];
  //    axpby(psi[0],src,src,0.0,coeff);
  //  }  else {
  //  }
  for(int s=0;s<nshift;s++) {
    axpby(psi[s],src,src,0.0,-bs[s]*alpha[s]);
  }
  

  // Iteration loop
  for (int k=1;k<=max_iter;k++){

    a = c /cp;
    axpy(p,p,r,a);

    // Note to self - direction ps is iterated seperately
    // for each shift. Does not appear to have any scope
    // for avoiding linear algebra in "single" case.
    // 
    // However SAME r is used. Could load "r" and update
    // ALL ps[s]. Bandwidth goes to that of "copy".
    // Could compare AXPY and COPY performance to estimate gain.
    for(int s=0;s<nshift;s++){
      if ( ! converged[s] ) {
	double as =a *z[s][iz]*bs[s] /(z[s][1-iz]*b);
	axpby(ps[s],r,ps[s],z[s][iz],as);
      }
    }

    cp=c;
    
    d = Mprec(p,mp,tmp,DaggerNo,1); 
        Mprec(mp,mmp,tmp,DaggerYes);

    bp=b;
    b=-cp/d;
    
    c=axpy_norm(r,mmp,r,b);

    // Toggle the recurrence history
    iz = 1-iz;
    for(int s=0;s<nshift;s++){
      if(!converged[s]){
	double z0 = z[s][1-iz];
	double z1 = z[s][iz];
	z[s][iz] = z0*z1*bp
	  / (b*a*(z1-z0) + z1*bp*(1- mass[s]*b)); 
	bs[s] = b*z[s][iz]/z0; // NB sign  rel to Mike
      }
    }
    
    // Fixme - Mike has a variable "at" that is missing here.
    // Understand this. NB sign of bs rel to Mike
    for(int s=0;s<nshift;s++){
      int ss = s;
      // Scope for optimisation here in case of "single".
      // Could load psi[0] and pull all ps[s] in.
      //      if ( single ) ss=primary;
      axpy(psi[ss],ps[s],psi[ss],-bs[s]*alpha[s]);
    }

    // Convergence checks
    int all_converged = 1;
    for(int s=0;s<nshift;s++){

      if (!converged[s]){

	double css  = c * z[s][iz]* z[s][iz];
	
	if(css<rsq[s]){
	  converged[s]=1;
	  if ( isBoss() && (!me) ) printf("bfmbase::CGNE_prec_multi: k=%d Shift %d has converged\n",k,s);
	} else {
	  all_converged=0;
	}

      }
    }

    if ( all_converged ){
      if ( isBoss() && (!me) )printf("bfmbase::CGNE_prec_multi: k=%d All shifts have converged\n",k);
      if ( isBoss() && (!me) )printf("bfmbase::CGNE_prec_multi: k=%d Checking solutions\n",k);
      // Check answers 
      for(int s=0; s < nshift; s++) { 
	Mprec(psi[s],mp,tmp,DaggerNo);
	Mprec(mp,mmp,tmp,DaggerYes);
        axpy(tmp,psi[s],mmp,mass[s]);
	axpy(r,tmp,src,-1);
	double rn = norm(r);
	double cn = norm(src);
	if ( isBoss() && !me ) {
	  printf("bfmbase::CGNE_prec_multi: shift[%d] true residual %le \n",s,sqrt(rn/cn));
	}
      }

      if ( single ) {
	for(int s=1; s < nshift; s++) { 
	  axpy(psi[0],psi[s],psi[0],1.0);
	}      
      }

      threadedFreeFermion(tmp);
      threadedFreeFermion(p);
      threadedFreeFermion(mp);
      threadedFreeFermion(mmp);
      threadedFreeFermion(r);
      for(int i=0;i<nshift;i++){
	threadedFreeFermion(ps[i]);
      }  

      return k;
    }
  }

  if ( isBoss() && !me ) printf("bfmbase::CGNE_prec_multi: CG not converged \n");

  threadedFreeFermion(tmp);
  threadedFreeFermion(p);
  threadedFreeFermion(mp);
  threadedFreeFermion(mmp);
  threadedFreeFermion(r);
  for(int i=0;i<nshift;i++){
    threadedFreeFermion(ps[i]);
  }  
}
#endif


#endif

