template <class Float>
int bfmbase<Float>::CGNE_G5D_M(Fermion_t solution[2], Fermion_t source[2])
{
  multi1d<bfm_fermion> evecs(0);
  multi1d<double> evals(0);
  return CGNE_G5D_M(solution,source, evecs,evals);
}
template <class Float>
int bfmbase<Float>::CGNE_G5D_M(Fermion_t solution[2], Fermion_t source[2], multi1d<bfm_fermion> &evecs, multi1d<double> &evals)
{
  int me = thread_barrier();
  Fermion_t src = threadedAllocFermion(); 
  Fermion_t hsrc = threadedAllocFermion(); 
  Fermion_t tmp = threadedAllocFermion(); 
  Fermion_t Mtmp= threadedAllocFermion(); 
  Fermion_t lsol= threadedAllocFermion(); 

  if (isBoss() && !me ) printf("bfmbase::CGNE: Solver type is %d\n",solver);

  double f;
  f = norm(source[0]);
  f+= norm(source[1]);

  if (isBoss() && !me ) printf("bfmbase::CGNE: Source norm is %le\n",f);
  f = norm(solution[0]);
  f+= norm(solution[1]);
  if (isBoss() && !me ) printf("bfmbase::CGNE: Guess norm is %le\n",f);

  // src_o = Mdag * (source_o - Moe MeeInv source_e)
  G5D_MooeeInv(source[Even],tmp,DaggerNo);
  f = norm(tmp);
  if (isBoss() && !me ) printf("bfmbase::CGNE: prec tmp norm is %le\n",f);
  G5D_Meo     (tmp,src,Odd,DaggerNo);
  f = norm(src);
  if (isBoss() && !me ) printf("bfmbase::CGNE: prec src norm is %le\n",f);
  axpy    (tmp,src,source[Odd],-1.0);
  f = norm(tmp);
  if (isBoss() && !me ) printf("bfmbase::CGNE: prec tmp norm is %le\n",f);
  G5D_Mprec(tmp,src,Mtmp,DaggerYes);  
  f = norm(src);
  if (isBoss() && !me ) printf("bfmbase::CGNE: prec src norm is %le\n",f);
  
  int Nev = evecs.size();
  if(Nev > 0){
    if ( isBoss() && !me ) printf("bfmbase::CGNE: deflating with %d evecs\n",evecs);
  }
  axpy    (hsrc,src,src,0.0);
  axpby   (lsol,src,src,0.0,0.0);
  
  for(int n = 0; n<Nev; n++){
   complex<double> cn = inner (evecs[n][1], src);
   ///deflate src
   axpy    (Mtmp,evecs[n][1],evecs[n][1],0.0);
   scale   (Mtmp,real(cn),imag(cn) );
   axpy    (hsrc,Mtmp,hsrc,-1.0);
   ///low solution
   axpy    (lsol, Mtmp, lsol, 1.0/evals[n]);
  }
  
  axpy(solution[Odd], hsrc, hsrc, 0.0);  
  int iter = CGNE_G5D_prec_MdagM(solution[Odd],src);
  axpy(solution[Odd], lsol, solution[Odd], 1.0);  
 
  threadedFreeFermion(Mtmp);
  threadedFreeFermion(lsol);
  threadedFreeFermion(hsrc);

  // sol_e = M_ee^-1 * ( src_e - Meo sol_o )...
  G5D_Meo(solution[Odd],tmp,Even,DaggerNo);
  axpy(src,tmp,source[Even],-1.0);
  G5D_MooeeInv(src,solution[Even],DaggerNo);
  
  f =norm(solution[0]);
  f+=norm(solution[1]);
  if (isBoss() && !me ) printf("bfmbase::CGNE: unprec sol norm is %le\n",f);
  
  threadedFreeFermion(tmp);
  threadedFreeFermion(src);

  return iter;
}

template <class Float>
int bfmbase<Float>::CGNE_G5D_prec_MdagM(Fermion_t psi, Fermion_t src)
{
  int iter;
  int me= thread_barrier();

  if ( reproduce ) { 

    uint64_t c1,c2;

    Fermion_t tmp_r = threadedAllocFermion(); 
    axpy    (tmp_r,psi,psi,0.0);

    c1=checksum(tmp_r);
    c2=checksum(psi);

    repro_control(ReproRecord);
    iter = CGNE_G5D_prec(tmp_r,src);
    if ( isBoss() && !me ) printf("bfmbase:: Checking reproducibility\n");
    repro_control(ReproCheck);
    iter = CGNE_G5D_prec(psi,src);
    repro_control(ReproIdle);

    c1=checksum(tmp_r);
    c2=checksum(psi);
    if ( c1!=c2  ) {
       printf("bfmbase:: result vector checksum miscompare in reproducing CG %16.16lx %16.16lx\n",c1,c2);
       exit(-1);
    }
    if ( isBoss() && !me ) printf("bfmbase:: reproducing cg passed -- checksums are %16.16lx %16.16lx\n",c1,c2);

    threadedFreeFermion(tmp_r);

  } else {

    iter = CGNE_G5D_prec(psi,src);
    repro_control(ReproIdle);

  }
  return iter;
}

template <class Float>
int bfmbase<Float>::CGNE_G5D_prec(Fermion_t psi, Fermion_t src)
{

  double f;
  double cp,c,a,d,b;
  int me = thread_barrier();

  Fermion_t p   = threadedAllocFermion(mem_fast); 
  Fermion_t tmp = threadedAllocFermion(mem_fast); 
  Fermion_t mp  = threadedAllocFermion(mem_fast); 
  Fermion_t mmp = threadedAllocFermion(mem_fast); 
  Fermion_t r   = threadedAllocFermion(mem_fast); 

  //Initial residual computation & set up

  G5D_Mprec(psi,mp,tmp,DaggerNo);
  G5D_Mprec(mp,mmp,tmp,DaggerYes);

  axpy (r, mmp, src,-1.0);
  axpy (p, mmp, src,-1.0);

  cp= norm(r);
  Float ssq =  norm(src);
  if ( verbose && isBoss() && !me ) {
    printf("bfmbase::CGNE_prec src %le \n",ssq);
  }
  Float rsq =  residual* residual*ssq;

  //Check if guess is really REALLY good :)
  if ( cp <= rsq ) {
    if ( verbose && isBoss() && !me ) {
      printf("bfmbase::CGNE_prec k=0 converged - suspiciously nice guess %le %le\n",cp,rsq);
    }
    threadedFreeFermion(tmp);
    threadedFreeFermion(p);
    threadedFreeFermion(mp);
    threadedFreeFermion(mmp);
    threadedFreeFermion(r);
    return 0;
  }
  if ( verbose && isBoss() && !me ) 
    printf("bfmbase::CGNE_prec k=0 residual %le rsq %le\n",cp,rsq);

  if ( isBoss() && !me ) {
    if ( this->watchfile ) {
      printf("bfmbase::CGNE_prec watching file \"%s\"\n",this->watchfile);
    }    
  }
  struct timeval start,stop;
  gettimeofday(&start,NULL);


  for (int k=1;k<=max_iter;k++){

    this->iter=k;
    uint64_t t_iter_1=GetTimeBase();

    c=cp;


    uint64_t t_mprec_1=GetTimeBase();
    d = G5D_Mprec(p,mp,tmp,0,1);
    uint64_t t_mprec_2=GetTimeBase();
    a = c/d;

    //NB: No longer need mp, overwrite
    //    With optimised version need to take
    //            mp-> Dmp
    //        Dmp,mp-> amp+DDmp -> mp inplace.
    // Implement "Mprec_inplace(mp,mmp,1); 
    uint64_t t_mprec_3=GetTimeBase();
    double qq=G5D_Mprec(mp,mmp,tmp,1); 
    uint64_t t_mprec_4=GetTimeBase();
    double b_pred = a*(a*qq-d)/c;

#if 0
    // Can fuse these together. 
    // Further, producing "p" in a cache-blocked fashion could allow no reload
    // of p from main memory (~8 timeslices at a time?)

    //Fused axpy-norm required.    
    //As know "a" in advance could save bandwidth by
    //not writing out mmp, but doing linear combination.
    //in Mprec.
    uint64_t tr1=GetTimeBase();
    cp = axpy_norm(r,mmp,r,-a);
    b = cp/c;
    uint64_t tr2=GetTimeBase();

    uint64_t tpsi1=GetTimeBase();
    axpy(psi,p,psi,a);
    uint64_t tpsi2=GetTimeBase();
    // New (conjugate/M-orthogonal) search direction
    uint64_t tp1=GetTimeBase();
    axpy(p,p,r,b);
    uint64_t tp2=GetTimeBase();
#else
    uint64_t tpsi1=GetTimeBase();
    uint64_t tp1=GetTimeBase();
    uint64_t tp2=tp1+1;
    uint64_t tr1=tp1;
    uint64_t tr2=tp2;
    b  = b_pred;
    cp = cg_psi_p(psi,p,r,mmp,a,b);
    uint64_t tpsi2=GetTimeBase();
#endif
    uint64_t t_iter_2=GetTimeBase();

    // Inefficient hack while I test something
    if(0){
      if ( this->isBoss() && !me ) {
	printf("k=%d alpha = %le\n",k,a);
	printf("k=%d predict beta=%le\n",k,b_pred);
	printf("k=%d beta = %le\n",k,b);
      }
    }

    
    // verbose nonsense
    if ( (this->iter==this->time_report_iter) && isBoss() && !me  ) {
      
      int lx = node_latt[0];
      int ly = node_latt[1];
      int lz = node_latt[2];
      int lt = node_latt[3];

      int cb4dsites = (lx*ly*lz*lt)/2;
      printf("fermionCacheFootprint: %ld \n",7*axpyBytes()/3);
      printf("gauge  CacheFootprint: %ld \n",2*18*8*cb4dsites*2);
      printf("fermionVecBytes      : %ld \n",axpyBytes()/3);
      printf("axpyBytes            : %ld \n",axpyBytes());
      printf("axpy      (soln)     : %ld cyc %le MB/s\n",(tpsi2-tpsi1),(double)axpyBytes()*1600./(tpsi2-tpsi1));
      printf("axpy_norm (residual) : %ld cyc %le MB/s\n",(tr2-tr1),(double)axpyBytes()*1600./(tr2-tr1));
      printf("axpy      (search)   : %ld cyc %le MB/s\n",(tp2-tp1),(double)axpyBytes()*1600./(tp2-tp1));
      printf("Iter time            : %ld cyc\n",t_iter_2-t_iter_1);
      printf("Mprec time           : %ld cyc\n",t_mprec_2-t_mprec_1);
      printf("Mprec time           : %ld cyc\n",t_mprec_4-t_mprec_3);
      fflush(stdout);
    }


    if ( ((k%100 == 0) && (verbose!=0)) || (verbose > 2)){
      if ( isBoss() && !me ) {
	printf("bfmbase::CGNE_prec: k=%d r^2=%le %le %lx\n",k,cp,sqrt(cp/ssq),this);
      }
      //      if ( this->CheckStopWatchFile() ) {
      //	printf("bfmbase::CGNE_prec: terminating due to stop request %s\n",this->watchfile);
      //abort();
      //      }
    }

    // Stopping condition
    if ( cp <= rsq ) { 

      gettimeofday(&stop,NULL);
      struct timeval diff;
      timersub(&stop,&start,&diff);

      if ( isBoss() && !me ) printf("bfmbase::CGNE_prec converged in %d iterations\n",k);
      if ( isBoss() && !me ) printf("bfmbase::CGNE_prec converged in %d.%6.6d s\n",diff.tv_sec,diff.tv_usec);


      double flops = mprecFlops()*2.0 + 2.0*axpyNormFlops() + axpyFlops()*2.0;
             flops = flops * k;

      double t = diff.tv_sec*1.0E6 + diff.tv_usec;
      if ( isBoss()&& !me ) 
       printf("bfmbase::CGNE_prec: %d mprec flops/site\n",mprecFlopsPerSite());
      if ( isBoss()&& !me ) printf("bfmbase::CGNE_prec: %le flops\n",flops);
      if ( isBoss()&& !me ) printf("bfmbase::CGNE_prec: %le mflops per node\n",flops/t);

      G5D_Mprec(psi,mp,tmp,0);
      G5D_Mprec(mp,mmp,tmp,1); 
      axpy(tmp,src,mmp,-1.0);
      double true_residual = sqrt(norm(tmp)/norm(src));
      if ( isBoss() && !me ) 
	printf("bfmbase::CGNE_prec: true residual is %le \n",true_residual);

      threadedFreeFermion(tmp);
      threadedFreeFermion(p);
      threadedFreeFermion(mp);
      threadedFreeFermion(mmp);
      threadedFreeFermion(r);
#ifdef LIST_ENGINE
      if ( this->list_engine) L1P_PatternUnconfigure();
#endif
      return k;
    }

  }
  if ( isBoss() && !me ) printf("bfmbase::CGNE_prec: CG not converged \n");
  threadedFreeFermion(tmp);
  threadedFreeFermion(p);
  threadedFreeFermion(mp);
  threadedFreeFermion(mmp);
  threadedFreeFermion(r);
#ifdef LIST_ENGINE
  if (this->list_engine ) L1P_PatternUnconfigure();
#endif
  return -1;

}

