#ifndef BFM_IMP_EX_H
#define BFM_IMP_EX_H
/*Peter Boyle April 2007*/
#include "bfm.h"
#include <omp.h>

#include <stdio.h>
#include <stdlib.h>


template <class Float>
void      bfmbase<Float>::precisionChange (Fermion_t in, Fermion_t out, int control, int cb)
{
  int Nspinco=12;

  int thrlen, throff;
  int work = node_latt[0]*node_latt[1]*node_latt[2]*node_latt[3];
  int me;

  this->thread_work(work,me,thrlen,throff);

  for (int site=0;site<thrlen;site++ ) { 
    
    int x[4] ;
    int si=site+throff;

    x[0]=si%node_latt[0];    si=si/node_latt[0];
    x[1]=si%node_latt[1];    si=si/node_latt[1];
    x[2]=si%node_latt[2];    si=si/node_latt[2];
    x[3]=si%node_latt[3];
    
    for (int s=0;s<this->Ls;s++ ) { 

      int sp;
      if ( precon_5d ) sp = s;
      else sp = 0;
    
      if ( ((x[0]+x[1]+x[2]+x[3]+sp)&0x1) == cb ) {      
	for ( int co=0;co<Nspinco;co++ ) { 
	  for ( int reim=0;reim<2;reim++ ) { 
	    double *dp;
	    float  *fp;
	    
	    int bidx = bagel_idx5d(x,s,reim,co,Nspinco,1);
	    
	    if ( control == SingleToDouble ) {
	      double *outp = (double *)out;
	      float  *inp  = (float *) in;
	      outp[bidx] = inp[bidx];
	    } else { 
	      double *inp  = (double *)in;
	      float  *outp = (float *) out;
	      outp[bidx] = inp[bidx];
	    }
	    
	  }
	}
      }   
    }}
  thread_barrier();
}

template <class Float>
void      bfmbase<Float>::importFermion  (QDPdouble *psi, Fermion_t handle,int cb)
{
  impexFermion(psi,handle,1,cb,0);
}
template <class Float>
void      bfmbase<Float>::exportFermion  (QDPdouble *psi, Fermion_t handle,int cb)
{
  impexFermion(psi,handle,0,cb,0);
}

template <class Float>
integer bfmbase<Float>::chroma_idx(int x[4],int reim,int i, int i_size)
{
  int ccb   = ((x[0]+x[1]+x[2]+x[3])&0x1); /*Work out local checkerboard of site*/ 
                         /*FIXME Here need to worry about base_parity
                          *as chroma indexes with global and not local parity
                          */
  int chroma_cbsite[4];
  chroma_cbsite[0] = x[0];
  chroma_cbsite[1] = x[1];
  chroma_cbsite[2] = x[2];
  chroma_cbsite[3] = x[3];

  int csite= x[0] + node_latt[0]*(x[1]
                  +       node_latt[1]*(x[2]
                  +             node_latt[2]*x[3]));
  csite = csite/2;
  int cbvol = (node_latt[0]*node_latt[1]*node_latt[2]*node_latt[3])/2;
  int idx = (ccb*cbvol+csite)*i_size*2 + i*2 + reim;

  return idx;  
}

// Double store the gauge field.
// Plan : do the cshift & adj in QDP, not in my code!
template <class Float>
void      bfmbase<Float>::importGauge(QDPdouble *gauge, int dir)
{
  int Ndircoco=72; /*8 directions stored*/
  int Ncoco=9;
  omp_set_num_threads(this->nthread);
#pragma omp parallel 
  {

#pragma omp for 
  for (int site=0;site<node_latt[0]*node_latt[1]*node_latt[2]*node_latt[3];site++ ) { 
    
    int x[4] ;
    int s=site;
    x[0]=s%node_latt[0];    s=s/node_latt[0];
    x[1]=s%node_latt[1];    s=s/node_latt[1];
    x[2]=s%node_latt[2];    s=s/node_latt[2];
    x[3]=s%node_latt[3];
    
      for ( int coco=0;coco<9;coco++ ) { 
      for ( int reim=0;reim<2;reim++ ) { 

        Float * bagel = this->u;
	int bbase = dir*9;
        int cidx = chroma_idx(x,reim,coco,Ncoco);
        int bidx = bagel_idx(x,reim,coco+bbase,Ndircoco,0);
        bagel[bidx] = gauge[cidx];


      }}
  }
  }
}
template <class Float>
void      bfmbase<Float>::impexFermion  (QDPdouble *psi, Fermion_t handle,int doimport,int cb,int s)
{
  int Nspinco=12;

  omp_set_num_threads(this->nthread);

#pragma omp parallel
  {

#pragma omp for 
  for (int site=0;site<node_latt[0]*node_latt[1]*node_latt[2]*node_latt[3];site++ ) { 
    
    int x[4] ;
    int si=site;
    x[0]=si%node_latt[0];    si=si/node_latt[0];
    x[1]=si%node_latt[1];    si=si/node_latt[1];
    x[2]=si%node_latt[2];    si=si/node_latt[2];
    x[3]=si%node_latt[3];
    
    int sp;
    if ( precon_5d ) sp = s;
    else sp = 0;

    if ( ((x[0]+x[1]+x[2]+x[3]+sp)&0x1) == cb ) {      
      for ( int co=0;co<Nspinco;co++ ) { 
      for ( int reim=0;reim<2;reim++ ) { 
        Float * bagel = (Float *)handle;
        int bidx = bagel_idx5d(x,s,reim,co,Nspinco,1);
        int cidx = chroma_idx(x,reim,co,Nspinco);
	if ( doimport ) bagel[bidx] = psi[cidx]; 
	else psi[cidx] = bagel[bidx] ;
      }}
    }      

  }
  }
}

template <class Float>
Fermion_t bfmbase<Float>::threadedAllocFermion   (int mem_type)
{
  Fermion_t ret;
  int me = thread_barrier();
  if ( me == 0 ) {
    ret = this->allocFermion(mem_type);
  }
  ret = this->thread_bcast(me,ret);
  thread_barrier();
  return ret;
}
template <class Float>
void      bfmbase<Float>::threadedFreeFermion    (Fermion_t handle)
{
  int me = thread_barrier();
  if ( me == 0 ) { 
    Float * ferm = ( Float * ) handle;
    bfm_free(ferm);
  }
  thread_barrier();
}

template <class Float>
Fermion_t bfmbase<Float>::allocFermion   (int mem_type)
{
  int words = 24 * nsimd * (simd_cbvol + simd_allbound*2);
  Float *ferm = (Float *)bfm_alloc(words*cbLs*sizeof(Float),mem_type);

  if(ferm == 0){
    if ( isBoss() ) printf("bfmbase::allocFermion\n");
    fflush(stdout);
    exit(-1);
  }
  return (Fermion_t)ferm;
}
template <class Float>
void      bfmbase<Float>::freeFermion    (Fermion_t handle)
{
  Float * ferm = ( Float * ) handle;
  bfm_free(ferm);
}
#endif

