#include <stdio.h>
#include <stdlib.h>
#include <qmp.h>



extern "C" { 
   void* memalign(size_t, size_t);
};

#include <bfm.h>
#include <bfmcommqmp.h>
#include <bfmcommspi.h>

//////////////////////////////////////////////////////////
// SPI thingy-me-bob's
// Save a 32MB region for our comms buffers??
//////////////////////////////////////////////////////////

#include <spi/include/mu/Descriptor.h>
#include <spi/include/mu/Descriptor_inlines.h>
#include <spi/include/mu/InjFifo.h>
#include <spi/include/mu/Addressing.h>
#include <spi/include/mu/Addressing_inlines.h>
#include <spi/include/mu/GIBarrier.h>
#include <spi/include/kernel/MU.h>
#include <hwi/include/bqc/classroute.h>
#include <spi/include/kernel/collective.h>
#include <firmware/include/personality.h>

#define NBRS (8)

/////////////////////////////////////////////////////
// Torus mapping related
/////////////////////////////////////////////////////
static uint64_t BfmSPI_Torus_FIFO_Map[10] = {
  MUHWI_DESCRIPTOR_TORUS_FIFO_MAP_AM,
  MUHWI_DESCRIPTOR_TORUS_FIFO_MAP_AP,
  MUHWI_DESCRIPTOR_TORUS_FIFO_MAP_BM,
  MUHWI_DESCRIPTOR_TORUS_FIFO_MAP_BP,
  MUHWI_DESCRIPTOR_TORUS_FIFO_MAP_CM,
  MUHWI_DESCRIPTOR_TORUS_FIFO_MAP_CP,
  MUHWI_DESCRIPTOR_TORUS_FIFO_MAP_DM,
  MUHWI_DESCRIPTOR_TORUS_FIFO_MAP_DP,
  MUHWI_DESCRIPTOR_TORUS_FIFO_MAP_EM,
  MUHWI_DESCRIPTOR_TORUS_FIFO_MAP_EP
};

static uint8_t BfmSPI_hints_abcd[10] = {
  MUHWI_PACKET_HINT_AM|MUHWI_PACKET_HINT_B_NONE|MUHWI_PACKET_HINT_C_NONE|MUHWI_PACKET_HINT_D_NONE,
  MUHWI_PACKET_HINT_AP|MUHWI_PACKET_HINT_B_NONE|MUHWI_PACKET_HINT_C_NONE|MUHWI_PACKET_HINT_D_NONE,
  MUHWI_PACKET_HINT_A_NONE|MUHWI_PACKET_HINT_BM|MUHWI_PACKET_HINT_C_NONE|MUHWI_PACKET_HINT_D_NONE,
  MUHWI_PACKET_HINT_A_NONE|MUHWI_PACKET_HINT_BP|MUHWI_PACKET_HINT_C_NONE|MUHWI_PACKET_HINT_D_NONE,
  MUHWI_PACKET_HINT_A_NONE|MUHWI_PACKET_HINT_B_NONE|MUHWI_PACKET_HINT_CM|MUHWI_PACKET_HINT_D_NONE,
  MUHWI_PACKET_HINT_A_NONE|MUHWI_PACKET_HINT_B_NONE|MUHWI_PACKET_HINT_CP|MUHWI_PACKET_HINT_D_NONE,
  MUHWI_PACKET_HINT_A_NONE|MUHWI_PACKET_HINT_B_NONE|MUHWI_PACKET_HINT_C_NONE|MUHWI_PACKET_HINT_DM,
  MUHWI_PACKET_HINT_A_NONE|MUHWI_PACKET_HINT_B_NONE|MUHWI_PACKET_HINT_C_NONE|MUHWI_PACKET_HINT_DP
};

static uint8_t BfmSPI_hints_e[10] = {
  MUHWI_PACKET_HINT_E_NONE|MUHWI_PACKET_DO_NOT_ROUTE_TO_IO_NODE|MUHWI_PACKET_USE_DETERMINISTIC_ROUTING| MUHWI_PACKET_DO_NOT_DEPOSIT,
  MUHWI_PACKET_HINT_E_NONE|MUHWI_PACKET_DO_NOT_ROUTE_TO_IO_NODE|MUHWI_PACKET_USE_DETERMINISTIC_ROUTING| MUHWI_PACKET_DO_NOT_DEPOSIT,  
  MUHWI_PACKET_HINT_E_NONE|MUHWI_PACKET_DO_NOT_ROUTE_TO_IO_NODE|MUHWI_PACKET_USE_DETERMINISTIC_ROUTING| MUHWI_PACKET_DO_NOT_DEPOSIT,  
  MUHWI_PACKET_HINT_E_NONE|MUHWI_PACKET_DO_NOT_ROUTE_TO_IO_NODE|MUHWI_PACKET_USE_DETERMINISTIC_ROUTING| MUHWI_PACKET_DO_NOT_DEPOSIT,  
  MUHWI_PACKET_HINT_E_NONE|MUHWI_PACKET_DO_NOT_ROUTE_TO_IO_NODE|MUHWI_PACKET_USE_DETERMINISTIC_ROUTING| MUHWI_PACKET_DO_NOT_DEPOSIT,  
  MUHWI_PACKET_HINT_E_NONE|MUHWI_PACKET_DO_NOT_ROUTE_TO_IO_NODE|MUHWI_PACKET_USE_DETERMINISTIC_ROUTING| MUHWI_PACKET_DO_NOT_DEPOSIT,  
  MUHWI_PACKET_HINT_E_NONE|MUHWI_PACKET_DO_NOT_ROUTE_TO_IO_NODE|MUHWI_PACKET_USE_DETERMINISTIC_ROUTING| MUHWI_PACKET_DO_NOT_DEPOSIT,  
  MUHWI_PACKET_HINT_E_NONE|MUHWI_PACKET_DO_NOT_ROUTE_TO_IO_NODE|MUHWI_PACKET_USE_DETERMINISTIC_ROUTING| MUHWI_PACKET_DO_NOT_DEPOSIT,  
  MUHWI_PACKET_HINT_EM    |MUHWI_PACKET_DO_NOT_ROUTE_TO_IO_NODE|MUHWI_PACKET_USE_DETERMINISTIC_ROUTING| MUHWI_PACKET_DO_NOT_DEPOSIT,  
  MUHWI_PACKET_HINT_EP    |MUHWI_PACKET_DO_NOT_ROUTE_TO_IO_NODE|MUHWI_PACKET_USE_DETERMINISTIC_ROUTING| MUHWI_PACKET_DO_NOT_DEPOSIT
};

extern "C" {
extern int                 QMP_Labcde[5];
extern MUHWI_Destination_t QMP_neigh[8];
extern MUHWI_Destination_t QMP_me;
}

//////////////////////////////////
// Fast barrier
//////////////////////////////////
static uint32_t BfmSPI_GIBarrierClassRoute = 0; // 0 defaults to comm_world
static MUSPI_GIBarrier_t BfmSPI_GIBarrier;

//////////////////////////////////
// Global reduction
//////////////////////////////////
static MUHWI_Descriptor_t *BfmSPI_gsum_descr;
static int BfmSPI_gsum_buf = 0;


/////////////////////////////////////////////////////////////////////////////
// Memory handling
/////////////////////////////////////////////////////////////////////////////
static uint64_t BFM_DEVICE_MEM_SIZE = (32*1024*1024) ;

#define CHECK(A) { if ( rc ) {printf(A); printf("Looks bad returned %d\n",rc); fflush(stdout); sleep(10) ; exit(rc);}}
//#define CHECK(A) { if ( this->isBoss() ) { printf(A); fflush(stdout);} if ( rc ) { exit(rc);}}

static int BfmSPI_MU_resources_initialised;
static int BfmSPI_MU_resources_reserved;

static Kernel_MemoryRegion_t BfmSPI_device_memory_region;
static unsigned char      *BfmSPI_device_memory;
static unsigned char      *BfmSPI_device_malloc_ptr;
static uint64_t            BfmSPI_device_bytes;

//Restore from these values when we init a BFM object
static unsigned char      *BfmSPI_device_malloc_ptr_save;
static unsigned char      *BfmSPI_device_bytes_save;

// Address translation
static uint64_t BfmSPI_BasePa;
static uint64_t BfmSPI_BaseVa;


static  Personality_t personality;
static int isBoss;

// It appears we must spread our allocation thinly over 
// different groups associated with cores. 
static int                              BfmSPI_sub_group[8] = {0,4,8,12,16,20,24,28};
static MUSPI_BaseAddressTableSubGroup_t BfmSPI_BAT_sub_group[8];
static uint32_t                         BfmSPI_BAT_id[8];
static MUSPI_InjFifoSubGroup_t          BfmSPI_fifo_sub_group[8];

static uint64_t            BfmSPI_HaloBytes[NBRS];

// Injection FIFO's (DMA senders)
static uint32_t                   BfmSPI_fifoids[NBRS] = {0,1,2,3,4,5,6,7};
static Kernel_InjFifoAttributes_t BfmSPI_injFifoAttrs[NBRS];


// Device memory resident buffers & reception counters
static void               *BfmSPI_send_mem;
static void               *BfmSPI_receive_mem;
static void               *BfmSPI_send_buffer[NBRS];
static void               *BfmSPI_receive_buffer[NBRS];
static void               *BfmSPI_simd_receive_buffer[2];
static volatile uint64_t  *BfmSPI_receive_counters;
// Put descriptors
static MUHWI_Descriptor_t *BfmSPI_descrs;
// What are these inj memory fifos????
#define INJ_MEMORY_FIFO_SIZE (2*1024)
static unsigned char      *BfmSPI_inj_memory_fifos[NBRS];

// Offsets from BasePA
static uint64_t  BfmSPI_offst_send_buffer[NBRS];
static uint64_t  BfmSPI_offst_receive_buffer[NBRS];
static uint64_t  BfmSPI_offst_receive_counters[NBRS];
static uint64_t  BfmSPI_offst_descr[NBRS];
static uint64_t  BfmSPI_offst_inj_memory_fifos[NBRS];
static uint      BfmSPI_HardwareThread;

void *comm_spi_malloc(size_t bytes)
{
  void * old = (void *)BfmSPI_device_malloc_ptr;
  BfmSPI_device_bytes     +=bytes;
  if ( BfmSPI_device_bytes > BFM_DEVICE_MEM_SIZE ) {
    printf("BAGEL: Out of SPI device memory\n");
    exit(-1);
  }
  BfmSPI_device_malloc_ptr+=bytes;
  return old;
}
void comm_spi_heap_restore(void)
{
  BfmSPI_device_malloc_ptr =BfmSPI_device_malloc_ptr_save;
  BfmSPI_device_bytes      =BfmSPI_device_bytes_save;
}
void comm_spi_heap_save(void)
{
  BfmSPI_device_malloc_ptr_save =BfmSPI_device_malloc_ptr;
  BfmSPI_device_bytes_save      =BfmSPI_device_bytes;
}

static int comm_spi_reserve = comm_spi_reserve_hardware(6);

int comm_spi_reserve_hardware(int log2mb) 
{
  if ( BfmSPI_MU_resources_reserved ) return 1;

  Kernel_GetPersonality(&personality, sizeof(personality));
  isBoss = 1;
  if ( personality.Network_Config.Acoord!=0 ) isBoss=0;
  if ( personality.Network_Config.Bcoord!=0 ) isBoss=0;
  if ( personality.Network_Config.Ccoord!=0 ) isBoss=0;
  if ( personality.Network_Config.Dcoord!=0 ) isBoss=0;
  if ( personality.Network_Config.Ecoord!=0 ) isBoss=0;

  uint64_t rc;
  BFM_DEVICE_MEM_SIZE = (32*1024*1024) ;
  if ( log2mb == 6 ) BFM_DEVICE_MEM_SIZE = ( 64*1024*1024) ;
  if ( log2mb == 7 ) BFM_DEVICE_MEM_SIZE = (128*1024*1024) ;
  if ( log2mb == 8 ) BFM_DEVICE_MEM_SIZE = (256*1024*1024) ;
  if ( log2mb == 9 ) BFM_DEVICE_MEM_SIZE = (512*1024*1024) ;
  if ( log2mb ==10 ) BFM_DEVICE_MEM_SIZE = (1024*1024*1024) ;
  if ( log2mb > 10 ) if(isBoss) printf("Warning truncating the device memory requested to 1GB so don't be greedy\n");

  if(isBoss) printf("bfmcommspi: Printing all sorts of ugly stuff to convince you this is hard\n");
  if(isBoss) printf("bfmcommspi: SPI device memory pool %ld bytes\n",BFM_DEVICE_MEM_SIZE);

  ////////////////////////////////////////////////
  // Device memory 
  ////////////////////////////////////////////////
  BfmSPI_HardwareThread    = Kernel_PhysicalHWThreadID();
  BfmSPI_device_memory     = (unsigned char *)memalign(BFM_DEVICE_MEM_SIZE,BFM_DEVICE_MEM_SIZE);
  BfmSPI_device_malloc_ptr = BfmSPI_device_memory;
  BfmSPI_device_bytes      = 0;
  
  if(isBoss) printf("bfmcommspi: allocated SPI device memory pool in virtual memory space 0x%16.16lx - 0x%16.16lx \n",BfmSPI_device_memory,
			       (integer)BfmSPI_device_memory+BFM_DEVICE_MEM_SIZE);

  rc = Kernel_CreateMemoryRegion (&BfmSPI_device_memory_region,
				   BfmSPI_device_memory,
				   BFM_DEVICE_MEM_SIZE);
  CHECK("Kernel_CreateMemoryRegion\n");
  if(isBoss) printf("bfmcommspi: called Kernel_CreateMemoryRegion to pin this in contiguous physical region\n");

  BfmSPI_BasePa = (uint64_t)BfmSPI_device_memory_region.BasePa;
  BfmSPI_BaseVa = (uint64_t)BfmSPI_device_memory_region.BaseVa;

  if(isBoss) printf("bfmcommspi: Memory translation is as follows:\n");
  if(isBoss) printf("bfmcommspi: SPI device memory Base Physical Address %16.16lx\n",BfmSPI_BasePa);
  if(isBoss) printf("bfmcommspi: SPI device memory Base Virtual  Address %16.16lx\n",BfmSPI_BaseVa);
  
  for(int g=0;g<NBRS;g++){

    BfmSPI_BAT_id[g] = 0;
    if(isBoss) printf("bfmcommspi: allocating Block Address Translation (BAT) id %d Group %d\n",BfmSPI_BAT_id[g],BfmSPI_sub_group[g]);

    rc = Kernel_AllocateBaseAddressTable (BfmSPI_sub_group[g],    // Thread/group
					 &BfmSPI_BAT_sub_group[g],// returned data struct
					  1,                   // 1 BAT id in list
					 &BfmSPI_BAT_id[g],    // Allocated entry
					  0 );                 // User use
    CHECK("Kernel_AllocateBaseAddressTable\n");

    rc = MUSPI_SetBaseAddress (&BfmSPI_BAT_sub_group[g],
			        BfmSPI_BAT_id[g],
			        BfmSPI_BasePa );
    CHECK("MUSPI_SetBaseAddress\n");
  }
  
  /////////////////////////////////////////////////////////////////
  // Place Receive Counters, descriptors and buffers in device mem
  /////////////////////////////////////////////////////////////////
  if(isBoss) printf("bfmcommspi: preallocating reception counters, buffers, descriptors in this device memory pool\n");

  BfmSPI_descrs           = (MUHWI_Descriptor_t *) comm_spi_malloc(sizeof(MUHWI_Descriptor)*NBRS);
  BfmSPI_gsum_descr       = (MUHWI_Descriptor_t *) comm_spi_malloc(sizeof(MUHWI_Descriptor));
  BfmSPI_receive_counters = (uint64_t *)           comm_spi_malloc(sizeof(uint64_t)*NBRS);

  //////////////////////////////////////////////////////////////////////
  // Injection FIFO's
  //////////////////////////////////////////////////////////////////////
  if(isBoss)   printf("bfmcommspi: Injection FIFO setup\n");

  for(int i=0;i<NBRS;i++){
    BfmSPI_injFifoAttrs[i].RemoteGet = 0;
    BfmSPI_injFifoAttrs[i].System    = 0;
  }

  for(int i=0;i<NBRS;i++){
    uint32_t numfifo = 1;
    rc = Kernel_AllocateInjFifos (BfmSPI_sub_group[i], 
				  &BfmSPI_fifo_sub_group[i],
				  numfifo, 
				  &BfmSPI_fifoids[i],
				  &BfmSPI_injFifoAttrs[i]);

    CHECK("Kernel_AllocateInjFifos\n");
    
  }

  for(int dir=0;dir<NBRS;dir++){

    BfmSPI_inj_memory_fifos[dir]= (unsigned char *)comm_spi_malloc(INJ_MEMORY_FIFO_SIZE);
    
    BfmSPI_offst_inj_memory_fifos[dir]= (uint64_t)BfmSPI_inj_memory_fifos[dir]-BfmSPI_BaseVa;
    BfmSPI_offst_descr[dir]           = (uint64_t)&BfmSPI_descrs[dir]         -BfmSPI_BaseVa;
    BfmSPI_offst_receive_counters[dir] = MUSPI_GetAtomicOffsetFromBaseAddress ( 
									    &BfmSPI_BAT_sub_group[dir],
									     BfmSPI_BAT_id[dir],
									     (uint64_t)&BfmSPI_receive_counters[dir]
									    -BfmSPI_BaseVa
									    +BfmSPI_BasePa,
									     MUHWI_ATOMIC_OPCODE_STORE_ADD );

    rc = Kernel_InjFifoInit (&BfmSPI_fifo_sub_group[dir],
			      BfmSPI_fifoids[dir],
			     &BfmSPI_device_memory_region,      
			      BfmSPI_offst_inj_memory_fifos[dir],
			      INJ_MEMORY_FIFO_SIZE-1);    
    CHECK("Kernel_InjFifoInit\n");
                
    rc = Kernel_InjFifoActivate (&BfmSPI_fifo_sub_group[dir], 1, &BfmSPI_fifoids[dir],
				 KERNEL_INJ_FIFO_ACTIVATE);    
    CHECK("Kernel_InjFifoActivate \n");

  }

  ///////////////////////////////////////////////////////////////
  // Fast barrier support -- MPI_Barrier appears to be "detuned"
  ///////////////////////////////////////////////////////////////
  rc = MUSPI_GIBarrierInit ( &BfmSPI_GIBarrier,BfmSPI_GIBarrierClassRoute);
  CHECK("MUSPI_GIBarrierInit\n");

  if(isBoss)   printf("bfmcommspi: Global Interrupt Barriers initialised\n");

  MUSPI_GIBarrierEnterAndWait ( &BfmSPI_GIBarrier );

  if(isBoss) printf("bfmcommspi: preallocated section of device heap state for reinitialisation\n");
  comm_spi_heap_save();

  if(isBoss) printf("bfmcommspi: SPI setup is complete\n");
  if(isBoss) printf("bfmcommspi: End of scary messages\n");
  fflush(stdout);

  MUSPI_GIBarrierEnterAndWait ( &BfmSPI_GIBarrier );

  return 0;
}

template <class Float> 
void bfmcommspi<Float>::comm_spi_barrier(void) 
{
  int rc = MUSPI_GIBarrierEnterAndWait ( &BfmSPI_GIBarrier );
  CHECK("MUSPI_GIBarrierEnterAndWait\n");
}

template <class Float> 
void bfmcommspi<Float>::comm_gsum(double &val)
{
  int me = thread_barrier();
  int rc;
  double myval=0;

  if ( me == 0 ) { 

    // Preload reception counter for 1 word gsum
    BfmSPI_receive_counters[BfmSPI_gsum_buf] = sizeof(double);
    *((double *)BfmSPI_receive_buffer[BfmSPI_gsum_buf]) = 0.0;    
    *((double *)BfmSPI_send_buffer[BfmSPI_gsum_buf])    = val;

    ppc_msync();  // make sure we pick up the correct start for this round
    
    // Might not be necessary?
    // comm_spi_barrier();

    // Send with SPI
    rc = MUSPI_InjFifoInject(MUSPI_IdToInjFifo( BfmSPI_fifoids[BfmSPI_gsum_buf],
						&BfmSPI_fifo_sub_group[BfmSPI_gsum_buf]),
			     BfmSPI_gsum_descr);
    if( rc == -1 ) { 
      printf("Gsum SPI FifoInject failed with code %d\n",rc);	fflush(stdout);
      exit(-1);
    }

    // Wait till done
    uint32_t polls = 0;
    while( BfmSPI_receive_counters[BfmSPI_gsum_buf]!= 0 ) { 
      polls++;
      if ( polls == 1024*1024 ) {
	polls=0;
      }
    }
    ppc_msync();  // make sure we pick up the correct start for this round
    myval = *((double *)BfmSPI_receive_buffer[BfmSPI_gsum_buf]);
  }

  thread_sum(myval,me); // Broadcast to all threads
  val = myval;
  
}

template <class Float> 
void bfmcommspi<Float>::comm_init(void) 
{

  const int A=0;
  const int B=1;
  const int C=2;
  const int D=3;
  const int E=4;

  
  /***************************
   *Reset the memory heap
   * - Warning coexistence of two bfms precludes single/double
   * - scope lock needed?
   ***************************
   */
  comm_spi_heap_restore();

  /*
   * Allocate device memory
   */
  uint64_t words = this->simd_allbound*SPINOR_SIZE*this->simd()*this->cbLs*2;
  uint64_t bytes = 2*words * sizeof(Float);
  uint64_t mem_pool = (bytes + 1024*1024);

  if ( mem_pool > BFM_DEVICE_MEM_SIZE ) { 
    printf("SPI Memory pool exceeds reserved memory pool\n");exit(0);
  }

  BfmSPI_send_mem    = comm_spi_malloc(words * sizeof(Float));
  BfmSPI_receive_mem = comm_spi_malloc(words * sizeof(Float));

  for(int mu=0;mu<4;mu++){
    for(int pm=0;pm<2;pm++){

      int dir= pm+2*mu;
      int simd_factor = this->simd();
      //      simd_factor = simd_factor/this->simd_grid[mu];
      int words = this->simd_nbound[mu] * HALF_SPINOR_SIZE * simd_factor * this->cbLs;

      BfmSPI_send_buffer[dir]    = &( (Float *)BfmSPI_send_mem)  [this->comm_offset[pm][mu]*SPINOR_SIZE*this->simd()*this->cbLs];
      BfmSPI_receive_buffer[dir] = &( (Float *)BfmSPI_receive_mem)[this->comm_offset[pm][mu]*SPINOR_SIZE*this->simd()*this->cbLs];

      if (this->simd_grid[mu] > 1 ){
	uint64_t simd_words = this->simd_nbound[mu] * SPINOR_SIZE * this->cbLs * 2;
	uint64_t simd_bytes = words * sizeof(Float);
	BfmSPI_simd_receive_buffer[pm]    = comm_spi_malloc(simd_bytes);
	BfmSPI_offst_receive_buffer[dir]  = (uint64_t)BfmSPI_simd_receive_buffer[pm]  -BfmSPI_BaseVa;
      } else {
	BfmSPI_offst_receive_buffer[dir]  = (uint64_t)BfmSPI_receive_buffer[dir]  -BfmSPI_BaseVa;
      }

      // Virt->Phys
      BfmSPI_offst_send_buffer[dir]     = (uint64_t)BfmSPI_send_buffer[dir]     -BfmSPI_BaseVa;

    }
  }

  for(int mu=0;mu<4;mu++){
    for(int pm=0;pm<2;pm++){

      int dir= pm+2*mu;
      int simd_factor = this->simd();
      simd_factor = simd_factor/this->simd_grid[mu];
      int words = this->simd_nbound[mu] * HALF_SPINOR_SIZE * simd_factor * this->cbLs;

      BfmSPI_HaloBytes[dir]       = words*sizeof(Float);      
      if (this->isBoss()) printf("bfmcommspi: Halo bytes[%d] = %d\n",dir,BfmSPI_HaloBytes[dir]);
      this->sendbufs[dir] = (Float *)BfmSPI_send_buffer[dir];
      if ( (this->simd_grid[mu] > 1) && !this->local_comm[mu]) {
	this->recvbufs[dir]  = (Float *)BfmSPI_receive_buffer[dir];
	this->simd_rbuf[dir] = (Float *)BfmSPI_simd_receive_buffer[pm];
      } else {
	this->recvbufs[dir] = (Float *)BfmSPI_receive_buffer[dir];
      }

    }
  }  

  ////////////////////////////////////////////////
  // DMA/FIFO controls
  ////////////////////////////////////////////////

  {
  MUSPI_Pt2PtDirectPutDescriptorInfo_t info;
  MUSPI_Pt2PtDescriptorInfoFields_t Pt2Pt;/**< Point-to-point descriptor info */
  MUSPI_DirectPutDescriptorInfoFields_t DirectPut;

  if (this->isBoss()) printf("bfmcommspi: DMA/FIFO controls\n");
  if (this->isBoss()) printf("bfmcommspi: Using Pt2PtDirectPut messages\n");
  if (this->isBoss()) printf("bfmcommspi: Using SPIGIBarrier() to synchronize\n");
  if (this->isBoss()) printf("bfmcommspi: Retrieving neigbour information from QMP and checking it matches hardware torus\n");
  if (this->isBoss()) printf("bfmcommspi: Must use PAB's QMP hack -qmp-geom native/native3d \n");

  for(int mu=0;mu<4;mu++){
    for(int pm=0;pm<2;pm++){

      int dir = pm+2*mu;
      int buf= (1-pm)+2*mu;

      int mach_dim = -1;
      int ndelta   = 0;
      int L=0;
      int me=0;
      int you=0;

      if ( QMP_neigh[dir].Destination.Destination != QMP_me.Destination.Destination ) {

	if ( QMP_neigh[dir].Destination.A_Destination != QMP_me.Destination.A_Destination ) {
	  mach_dim=A; ndelta++;
	}
	if ( QMP_neigh[dir].Destination.B_Destination != QMP_me.Destination.B_Destination ) {
	  mach_dim=B; ndelta++;
	}
	if ( QMP_neigh[dir].Destination.C_Destination != QMP_me.Destination.C_Destination ) {
	  mach_dim=C; ndelta++;
	}
	if ( QMP_neigh[dir].Destination.D_Destination != QMP_me.Destination.D_Destination ) {
	  mach_dim=D; ndelta++;
	}
	if ( QMP_neigh[dir].Destination.E_Destination != QMP_me.Destination.E_Destination ) {
	  mach_dim=E; ndelta++;
	}
	if ( ndelta!= 1) { 
	  printf("bfmspi: mapping logic bomb 2\n");
	}

	switch(mach_dim){ 
	case A:
	  me =QMP_me.Destination.A_Destination;
	  you=QMP_neigh[dir].Destination.A_Destination;
	  L  =QMP_Labcde[A];
	  break;
	case B:
	  me =QMP_me.Destination.B_Destination;
	  you=QMP_neigh[dir].Destination.B_Destination;
	  L  =QMP_Labcde[B];
	  break;
	case C:
	  me =QMP_me.Destination.C_Destination;
	  you=QMP_neigh[dir].Destination.C_Destination;
	  L  =QMP_Labcde[C];
	  break;
	case D:
	  me =QMP_me.Destination.D_Destination;
	  you=QMP_neigh[dir].Destination.D_Destination;
	  L  =QMP_Labcde[D];
	  break;
	case E:
	  me =QMP_me.Destination.E_Destination;
	  you=QMP_neigh[dir].Destination.E_Destination;
	  L  =QMP_Labcde[E];
	  break;
	default:
	  printf("bfmspi: mapping logic bomb 3\n");
	  exit(-1);
	}
	
	int mach_pm=-1;
	if ( me == ((you+1)) ) { 
	  mach_pm=0;
	} else if ( you = ((me+1)) ){
	  mach_pm=1;
	} else if ( me == ((you+1)%L) ) { // Avoid peri wrap if possible
	  mach_pm=0;
	} else if ( you = ((me+1)%L) ){
	  mach_pm=1;
	} else { 
	  printf("bfmspi: mapping logic bomb 4\n");
	  exit(-1);
	}

	  
 	int mach_dir = mach_dim*2 + mach_pm;

        uint64_t flagbit = FW_BIT(mach_dir);
	if ( !(personality.Network_Config.NetFlags2 & flagbit) ) {
	  printf ("machdir %d is not enabled\n",mach_dir);
	}
	
	info.Base.Pre_Fetch_Only = 0;
	info.Base.Payload_Address= BfmSPI_offst_send_buffer[buf]+BfmSPI_BasePa; // phys addr
	info.Base.Message_Length = BfmSPI_HaloBytes[buf];  
	info.Base.Torus_FIFO_Map = BfmSPI_Torus_FIFO_Map[mach_dir];
	info.Base.Dest           = QMP_neigh[dir];

	info.Pt2Pt.Hints_ABCD = BfmSPI_hints_abcd[mach_dir];
	info.Pt2Pt.Misc1 = BfmSPI_hints_e[mach_dir];
	info.Pt2Pt.Misc2 = MUHWI_PACKET_VIRTUAL_CHANNEL_DETERMINISTIC;
	info.Pt2Pt.Skip  = 8;

	info.DirectPut.Rec_Payload_Base_Address_Id = BfmSPI_BAT_id[dir];
	info.DirectPut.Rec_Payload_Offset          = BfmSPI_offst_receive_buffer[buf];
	info.DirectPut.Rec_Counter_Base_Address_Id = BfmSPI_BAT_id[dir];
	info.DirectPut.Rec_Counter_Offset          = BfmSPI_offst_receive_counters[0];
	info.DirectPut.Pacing                      = MUHWI_PACKET_DIRECT_PUT_IS_NOT_PACED;

 	int rc = MUSPI_CreatePt2PtDirectPutDescriptor(&BfmSPI_descrs[dir],&info);
	CHECK("MUSPI_CreatePt2PtDirectPutDescriptor \n");

      } else {
	if ( !this->local_comm[mu] ) {
	  printf("bfmspi: mapping logic bomb 1\n");
	  exit(-1);
	}
      }
    }
  }
  }

  if (this->isBoss()) printf("bfmcommspi: Created all SPI Pt2PtDirectPut descriptors \n");

  ///////////////////////////////////////////
  // Fast global sum
  ///////////////////////////////////////////
  {
    
    MUHWI_Destination_t dest;
    MUSPI_SetUpDestination ( &dest, 0, 0, 0, 0, 0 );

    MUSPI_CollectiveDirectPutDescriptorInfo_t  info;

    // Reuse send buffer [0] and FIFO_Map[0] for the send
    info.Base.Pre_Fetch_Only = 0;
    info.Base.Payload_Address= BfmSPI_offst_send_buffer[BfmSPI_gsum_buf]+BfmSPI_BasePa; // phys addr
    info.Base.Message_Length = sizeof(double);
    info.Base.Torus_FIFO_Map = MUHWI_DESCRIPTOR_TORUS_FIFO_MAP_CUSER;
    info.Base.Dest           = dest;

    info.DirectPut.Rec_Payload_Base_Address_Id = BfmSPI_BAT_id[BfmSPI_gsum_buf];
    info.DirectPut.Rec_Payload_Offset          = BfmSPI_offst_receive_buffer[BfmSPI_gsum_buf];
    info.DirectPut.Rec_Counter_Base_Address_Id = BfmSPI_BAT_id[BfmSPI_gsum_buf];
    info.DirectPut.Rec_Counter_Offset          = BfmSPI_offst_receive_counters[BfmSPI_gsum_buf];
    info.DirectPut.Pacing                      = MUHWI_PACKET_DIRECT_PUT_IS_NOT_PACED;
  
    info.Collective.Class_Route = 0;
    info.Collective.Word_Length = sizeof(double);
    info.Collective.Op_Code     = MUHWI_COLLECTIVE_OP_CODE_FLOATING_POINT_ADD;
    info.Collective.Misc        = MUHWI_PACKET_VIRTUAL_CHANNEL_USER_SUB_COMM | MUHWI_COLLECTIVE_TYPE_ALLREDUCE;
    info.Collective.Skip        = 0;

    int rc = MUSPI_CreateCollectiveDirectPutDescriptor(BfmSPI_gsum_descr,&info);
    CHECK("MUSPI_CreateCollectiveDirectPutDescriptor \n");


  }
  

}
template <class Float> 
void bfmcommspi<Float>::comm_end(void) 
{
  // We simply reinit everything in static allocation
}

template <class Float> 
void bfmcommspi<Float>::comm_start(int result_cb, Fermion_t psi,int dag) 
{
  // gather the faces. Routines here are threaded.
  // All threads cooperate in gathering.
  int me = this->thread_barrier();
  int t1=GetTimeBase();
  gather (result_cb,psi,dag);
  this->thread_barrier();
  int t2=GetTimeBase();

  if ( me == 0 ) {

    comm_spi_start();
    int t3=GetTimeBase();
    if(this->iter == this->time_report_iter+2){ 
      if ( this->isBoss() ) {
	printf("gather    : %d cyc\n",t2-t1);
	printf("spi_start : %d cyc\n",t3-t2);
      }
    }    
  }
  this->thread_barrier();
  
  return;
}

template <class Float> 
void bfmcommspi<Float>::comm_complete(int result_cb, Fermion_t psi) 
{
  int me=this->thread_barrier();
  if ( me == 0 ) {
    comm_spi_complete() ;
  }
  this->thread_barrier();
  this->comm_merge();
  this->thread_barrier();

  return;
}

template <class Float> 
void bfmcommspi<Float>::comm_spi_start(void) 
{
  BfmSPI_receive_counters[0]=0;
  int incr=0;

  uint64_t t0=GetTimeBase();
  this->comm_spi_barrier();

  uint64_t t1=GetTimeBase();
  for(int mu=0;mu<4;mu++){
    for(int pm=0;pm<2;pm++){

      int dir= pm+2*mu;
      if ( !this->local_comm[mu] ) {
	incr+=BfmSPI_HaloBytes[dir];  
      }
    }
  }

  BfmSPI_receive_counters[0]=incr;
  ppc_msync();

  uint64_t t2=GetTimeBase();
  this->comm_spi_barrier();

  uint64_t t3=GetTimeBase();


  for(int mu=0;mu<4;mu++){
    for(int pm=0;pm<2;pm++){

      if ( !this->local_comm[mu] ) {
	int dir= pm+2*mu;
	//MUSPI_DescriptorDumpHex("Before Injecting...",&BfmSPI_descrs[dir]);fflush(stdout);
	uint64_t rc = MUSPI_InjFifoInject(MUSPI_IdToInjFifo( BfmSPI_fifoids[dir],&BfmSPI_fifo_sub_group[dir]),&BfmSPI_descrs[dir]);
	if ( rc == ((uint64_t)-1) ) {
	  printf("MUSPI_InjFifoInject\n");
	  exit(-1);
	}
      }
    }
  }
  uint64_t t4=GetTimeBase();

    if(this->iter == this->time_report_iter-20){ 
      if ( this->isBoss() ) {
	printf("barrier   : %d cyc\n",t1-t0);
	printf("counters  : %d cyc\n",t2-t1);
	printf("barrier   : %d cyc\n",t3-t2);
	printf("inject    : %d cyc\n",t4-t3);
      }
    }    
  
}

template <class Float> 
void bfmcommspi<Float>::comm_spi_complete(void) 
{
  uint64_t old = 0;
  uint64_t stale=0;
  uint64_t polls=0;

  while( BfmSPI_receive_counters[0]!= 0 ) { 
    polls++;
    if ( polls == 1024*1024*1024 ) {
      polls=0;
      printf("Inordinate spinning in comms receive\n");fflush(stdout);
    }
  }
  ppc_msync();
}

template class bfmcommspi <float>;
template class bfmcommspi <double>;
 
