#ifndef KRYLOV_H
#define KRYLOV_H

#include <omp.h>
#include "Deflate.h"
#include <cstdlib>
#include <string>
#include <stdexcept>
#include <cmath>
#include <vector>
#include <algorithm>
#include <iostream>
#include <time.h>


/** Different ways to sort **/
template <class T> inline bool cf_large(T i, T j) { return (abs(i) < abs(j)); }
template <class T> inline bool cf_small(T i, T j) { return (abs(i) > abs(j)); }

/** Find the largest/smallest eigenvalues **/
enum ordertype {small, large};	
/** Doing arnoldi or lanczos ? **/
enum krylovtype {arn, lan};

/** Check to see if element 'test' is in the vector 'converged' **/ //TODO OBSOLETE
template <class T> inline bool isConv(vector<T> converged, T test, T c){
for(int i=0;i<converged.size();i++){if( abs(converged[i] - test) < c){return true;}}
return false;}

/** Check to see if element 'test' better (smaller/larger) than any element of 'converged' **/ //TODO OBSOLETE
template <class T> inline int isBetter(vector<T> converged, T test, ordertype comp){

vector<complex<T> > ord = converged;
bool better = false;
if(comp == small){ sort(converged.begin(),converged.end(),cf_small); if( cf_small(converged[0],test) ){better = true;} }
if(comp == large){ sort(converged.begin(),converged.end(),cf_large); if( cf_large(converged[0],test) ){better = true;} }

if(better){

	double diff = abs(ord[0]-converged[0]);
	int x=0;
	for(int i=0;i<converged.size();i++){
		double tdiff = abs(ord[i]-converged[0]);
		if(tdiff < diff){x = i; diff = tdiff;}
	}
	return x;
	}
else{	
	return -1;
	}
}


/** A sorting algorithm so bad it makes small children cry.
    keeps track of which eval goes with which evec		**/
template <class T> void inline EigenSort(vector<T>  &evals, vector<vector<T> > &evecs, ordertype comp){

int N = evals.size();
vector<T> unsorted(N); unsorted = evals;

if(comp == small) sort(evals.begin(),evals.end(),cf_small<T>);
if(comp == large) sort(evals.begin(),evals.end(),cf_large<T>);

vector<vector<T> > sortedvecs(N);
vector<int> derange(N);

for(int j=0;j<N;j++){
		double min = abs(evals[j] - unsorted[0]); derange[j] = 0;
		for(int i=1;i<N;i++){
			double tmin = abs(evals[j] - unsorted[i]);
		if(tmin < min){derange[j] = i; min = tmin;}	
		}
}

for(int i=0;i<N;i++){sortedvecs[i] = evecs[derange[i]];}
evecs = sortedvecs;
}


template <class T> void EigenSort(multi1d<Complex> &vals, multi1d<LatticeFermion> &evecs){

int N = vals.size();
vector<T> evals(N); for(int i=0;i<N;i++){evals[i] = toDouble( real(vals[i]) );}
vector<T> unsorted(N); unsorted = evals;

sort(evals.begin(),evals.end(),cf_small<T>);

multi1d<LatticeFermion> sortedvecs(N);
vector<int> derange(N);

for(int j=0;j<N;j++){
		double min = abs(evals[j] - unsorted[0]); derange[j] = 0;
		for(int i=1;i<N;i++){
			double tmin = abs(evals[j] - unsorted[i]);
		if(tmin < min){derange[j] = i; min = tmin;}	
		}
}

for(int i=0;i<N;i++){sortedvecs[i] = evecs[derange[i]];}
for(int i=0;i<N;i++){evecs[i] = sortedvecs[i];}

multi1d<Complex> sorted(N);
for(int i=0;i<N;i++){sorted[i] = vals[derange[i]];}
for(int i=0;i<N;i++){vals[i] = sorted[N-i-1];}

}


template <class T> void EigenSort(multi1d<Complex> &vals, multi1d<multi1d<LatticeFermion> > &evecs){

int N = vals.size();
int Ls = evecs[0].size();

vector<T> evals(N); for(int i=0;i<N;i++){evals[i] = toDouble( real(vals[i]) );}
vector<T> unsorted(N); unsorted = evals;

sort(evals.begin(),evals.end(),cf_small<T>);

multi1d<multi1d<LatticeFermion> > sortedvecs(N); for(int j=0;j<N;j++){sortedvecs[j].resize(Ls);}
vector<int> derange(N);

for(int j=0;j<N;j++){
		double min = abs(evals[j] - unsorted[0]); derange[j] = 0;
		for(int i=1;i<N;i++){
			double tmin = abs(evals[j] - unsorted[i]);
		if(tmin < min){derange[j] = i; min = tmin;}	
		}
}

for(int i=0;i<N;i++){
  for(int s = 0;s<Ls;s++){
  sortedvecs[i][s] = evecs[derange[i]][s];}}
for(int i=0;i<N;i++){
  for(int s = 0;s<Ls;s++){evecs[i][s] = sortedvecs[i][s];}}

multi1d<Complex> sorted(N);
for(int i=0;i<N;i++){sorted[i] = vals[derange[i]];}
for(int i=0;i<N;i++){vals[i] = sorted[N-i-1];}

}

enum EigenType {D, DDAG, G5D, DDAGD}; 

/**
An abstract iterative Krylov eigensolver.
**/
template <class S> 
class Krylov : public bfm_internal<S> {

public:

  int M;  ///Dimension M Krylov space
  int K;  ///Want K converged vectors
  int P;  ///Num shifts
  
  int con; ///number converged so far
  int maxits; ///maxiterations
  bool stop;	///converged
  double conv;  ///Convergence
  double convQR;	///convergence of intermediate QR solves
  ordertype comp; ///Type of evals wanted
  Matrix<S> H;	///Hessenberg matrix
  vector<S> converged; ///converged evals
  krylovtype kr; ///Lanczos or Arnoldi
  int mvprod; ///Counts matrix vector products
  int innerprod; ///Counts inner products
  bool lock; ///Use locking transofrmation or not
  int iterations;///counts its
  bool cont;	///Continuation of pervious calc
  bool refine;	///Improve previous calc
  EigenType D;	///The operator type

  double rho;   ///orthagonality requirement for Lanczos
  int ref;      ///orthagonality requirement for Lanczos

  int prec; ///preconditioned = 1 or not = 0
  int prectype; ///5d = 1, 4d = 0;
  
  int nwils; ///for strange wilson mass term
  int save5d; ///How many 5d slices of the eigenvector to save
  
  int Tn;	///Order of Chebyshev polynomial
  bool sh; 	///Shifting or not
  double ch_rad; ///Spectral radius
  double ch_off; ///Spectral offset (ie. find eigenvalues of magnitude less than this)
  double ch_shift; ///Shift the peak
  
  int get;
  multi1d<S> shift_extra;
  
  ///dwf operator args
  bfmarg dwfa;
  bfm_qdp<S> dop;
  
  ///lattice dimensions
  int lx, ly, lz, lt, Ls, gauge;
  
  ///DWF mass parameters
  Real M5;
  Real mq;

  ///gauge field
  multi1d<LatticeColorMatrix>  u; 
  
  ///Basis and residual
  multi1d<bfm_fermion> bq;
  multi1d<bfm_fermion> bf;
  multi1d<S> bl;

  ///Evals and evecs of Hessenberg/Tridiagonal matrix
  vector<S> evals;
  vector<vector<S> > evecs;

  Krylov(bfm_qdp<S> &dwf);
  ~Krylov();
  void Run();
  void ObliqueProj(Fermion_t b[2], Fermion_t x[2]);

  protected:
  ///Useful temporaries
  Fermion_t tmp1[2], tmp2[2]; 
  Fermion_t tmp;
  bool initialized;


  void Krylov_init();
  void dwfa_setup(bfm_qdp<S> &dwf);
  virtual void init() =0;
  void Check();
  void ImplicitRestart(	int TM, vector<S> &evals,  vector<vector<S> > &evecs);
  void TestConv(int SS, vector<S> &tevals, vector<vector<S> > &tevecs);
  void True_Residual( void );
  int Lanczos_Factor(int start, int end);
  int Arnoldi_Factor(int start, int end);
  void G5R(Fermion_t result[2], Fermion_t input[2]);
  void Abort(int ff, vector<S> &evals,  vector<vector<S> > &evecs);
  void Shift(Matrix<S> &Q, vector<S> shifts);

  virtual void herm_mult(bfm_fermion input, bfm_fermion &result) =0;
  virtual void dwf_multiply(bfm_fermion input, bfm_fermion &result) =0;
  void Gram(Fermion_t r[2], multi1d<bfm_fermion> &q, int N);
  void times_real(multi1d<bfm_fermion> &q, Matrix<S> &Q, int N);
  void times(multi1d<bfm_fermion> &q, Matrix<S> &Q, int N);
  void axpy(Fermion_t r[2], double a, Fermion_t x[2], Fermion_t y[2]);
  double axpy_norm(Fermion_t r[2], double a, Fermion_t x[2], Fermion_t y[2]);
  void axpby(Fermion_t r[2], double a, Fermion_t x[2],double b, Fermion_t y[2]);
  double axpby_norm(Fermion_t r[2], double a, Fermion_t x[2], double b, Fermion_t y[2]);
  void bfm_to_qdp(Fermion_t in[2], multi1d<LatticeFermion> &out);
  void bfm_to_qdp(Fermion_t in[2], LatticeFermion &out);
  void qdp_to_bfm(multi1d<LatticeFermion> &in, Fermion_t out[2]);
  void qdp_to_bfm(LatticeFermion in, Fermion_t out[2]);
  void free_fermion(Fermion_t x[2]);
  void zero_fermion(Fermion_t x[2]);
  void init_fermion(Fermion_t x[2]);
  void equate(Fermion_t x[2], Fermion_t y[2] );
  void scale(Fermion_t x[2], S s); 
  void scale(Fermion_t x[2], complex<double> s); 
  double norm(Fermion_t x[2]);
  double norm2(Fermion_t x[2]);
  void gamma5(Fermion_t x[2]); 
  double innerProduct_real(Fermion_t x[2], Fermion_t y[2] );
  complex<double> innerProduct(Fermion_t x[2], Fermion_t y[2] );
  double Chebyshev(double x);
  void Chebyshev(bfm_fermion input, bfm_fermion &v1);
  double qpoly(double alpha, double beta, double x);
  void qpoly(double beta, double alpha, bfm_fermion input, bfm_fermion &result);
  void shift_multiply(bfm_fermion input, bfm_fermion &result);
  void multiply(bfm_fermion input, bfm_fermion &result);

};



  /// Constructor
  template <class S> Krylov<S>::Krylov(bfm_qdp<S> &dwf){

    con = 0;
    conv = 1e-10;	//TODO residual underestimates true error, converge more accurately, usually fine.
    convQR = 1e-14;
    maxits = 10000;
    comp = small;
    stop = false;
    srand48(123456789);
    mvprod = 0;
    innerprod = 0;
    lock = true;
    iterations = 0;
    cont = false;
    refine = false;
    initialized = false;
    rho = 1.0;
    ref=0;

    nwils = 0;
    shift_extra.resize(0);
    lx = dwf.node_latt[0]; ly = dwf.node_latt[1]; lz = dwf.node_latt[2]; lt = dwf.node_latt[3]; 
    Ls = dwf.Ls; 
    M5 = dwf.M5;
    mq = dwf.mass;

    save5d = Ls;
    sh = false;
    
    dwfa_setup(dwf);

  }

  ///Basic initialization
  template <class S> void Krylov<S>::Krylov_init(){

    this->P = this->M-this->K;
    this->H.resize(this->M); this->H.Fill(0.0);
    this->evals.resize(this->M);
    this->evecs.resize(this->M);
    if(this->sh) this->ch_rad += this->ch_shift;
    this->conv = this->conv/100.0;
    this->get = this->K;
    (this->Tn > 0) ? this->comp = large : this->comp = small;
  }

  /// Setup DWF operator
  template <class S> void Krylov<S>::dwfa_setup(bfm_qdp<S> &dwf){

    dwfa.node_latt[0]  = lx;
    dwfa.node_latt[1]  = ly;
    dwfa.node_latt[2]  = lz;
    dwfa.node_latt[3]  = lt;

    multi1d<int> procs = QDP::Layout::logicalSize();
    QDPIO::cout  << "Krylov: machine dimension "<< procs.size()<<"(";
    for(int mu=0;mu<4;mu++){
      QDPIO::cout << procs[mu] << " ";
      if ( procs[mu]>1 ) {
	dwfa.local_comm[mu] = 0;
      } else { 
	dwfa.local_comm[mu] = 1;
      }
    }
    QDPIO::cout << ")"<< endl;
    QDPIO::cout << "Krylov: Local comm = " ;
    for(int mu=0;mu<4;mu++){
      QDPIO::cout << dwfa.local_comm[mu] << " ";
    }
    QDPIO::cout << endl;
    QDPIO::cout << "Krylov: Threads = " <<dwfa.threads << endl;
    
    multi1d<int> ncoor = QDP::Layout::nodeCoord();
    
    dwfa.solver = dwf.solver;
    dwfa.Ls   = Ls;
    dwfa.M5   = toDouble(M5);
    dwfa.mass = toDouble(mq);
    dwfa.Csw  = dwf.Csw;
    dwfa.max_iter = dwf.max_iter;

    QDPIO::cout << "Converging to precision " << conv << endl;  
    dwfa.residual = conv/100.0;  ///Do PV inverse to this accuracy

    u.resize(Nd);

  }

  ///Destructor
  template <class S> Krylov<S>::~Krylov(){

	dop.threadedFreeFermion(tmp);
	this->free_fermion(tmp1);
	this->free_fermion(tmp2);
	for(int i=0;i<bq.size();i++){this->free_fermion(bq[i]);}
	for(int i=0;i<bf.size();i++){this->free_fermion(bf[i]);}
	dop.end();

  }



  ///Run the Eigensolver
  template <class S> void Krylov<S>::Run(){
  init();
  int ff = this->con;

  QDPIO::cout << "Krylov: Run()"<<endl;

  if(this->kr == lan) {
    QDPIO::cout << "Krylov: Lanczos_Factor() ff = "<< ff << " " << this->M << endl;
    ff = Lanczos_Factor(ff,this->M);
  }
  /*if(this->kr == arn) {
    QDPIO::cout << "Krylov: Arnoldi_Factor()"<<endl;
    ff = Arnoldi_Factor(ff,this->M);
  }*/
  QDPIO::cout << "Krylov: ff "<<ff <<endl;
  
  if(ff < this->M){
    QDPIO::cout << "Krylov: aborting ff "<<ff <<" "<<this->M<<endl;
    Abort(ff,this->evals,this->evecs);
  }
  
  int thiscon = this->con;
  int itcount = 0;
  bool stop = false;

  
  for(int it=0;it<this->maxits && (this->con < this->get);it++){

    this->iterations++;
    itcount++;	
    QDPIO::cout << "Krylov: Iteration --> " << it << endl;
    std::vector<S> tevals(this->M-this->con);
    std::vector<std::vector<S> > tevecs(this->M-this->con);
    
    //True_Residual();
    TestConv(this->M, tevals, tevecs);
    if(this->con >= this->get){break;}
    ImplicitRestart(ff, tevals,tevecs);
  }

  /*if(this->kr == arn){ 
    QDPIO::cout << "Krylov: Arnoldi QReigensystem"<<endl;
    Eigensystem<S>(this->H, this->evals, this->evecs, this->convQR);
  }*/
  if(this->kr == lan){ 
    SymmEigensystem<S>(this->H, this->evals, this->evecs, this->convQR);
  }
  Check();
  QDPIO::cout << "Done" << endl;

  }

  ///H - shift I = QR; H = Q* H Q
  template <class S> void Krylov<S>::Shift(Matrix<S> &Q, vector<S> shifts){

  int P = shifts.size();
  int M = Q.dim;
  Q.Unity();
  
  for(int i=0;i<P;i++){
    
    S x, y, z;
    vector<S> ck(3), v(3);
    
    x = H(con+0,con+0)-shifts[i];
    y = H(con+1,con+0);
    ck[0] = x; ck[1] = y; ck[2] = 0; 
    
    normalize(ck);	///Normalization cancels in PHP anyway
    S beta;
    
    Householder_vector<S>(ck, 0, 2, v, beta);
    Householder_mult<S>(H,v,beta,0,con+0,con+2,0);
    Householder_mult<S>(H,v,beta,0,con+0,con+2,1);
    ///Accumulate eigenvector
    Householder_mult<S>(Q,v,beta,0,con+0,con+2,1);
    int sw = 0;
    
    for(int k=con+0;k<M-2;k++){

      x = H(k+1,k); 
      y = H(k+2,k); 
      z = (S)0.0;
      if(k+3 <= M-1){
	z = H(k+3,k);
      }else{
	sw = 1; v[2] = 0.0;
      }

      ck[0] = x; ck[1] = y; ck[2] = z;
	
      normalize(ck);
      
      Householder_vector<S>(ck, 0, 2-sw, v, beta);
      Householder_mult<S>(H,v, beta,0,k+1,k+3-sw,0);
      Householder_mult<S>(H,v, beta,0,k+1,k+3-sw,1);
      ///Accumulate eigenvector
      Householder_mult<S>(Q,v, beta,0,k+1,k+3-sw,1);
    }
    }
  }


  
  ///Only for Lanczos! use the one above for Arnoldi!!
  ///TODO Reverted to above, some bug when K = 1
  /*
  template <class S> void Krylov<S>::Shift(Matrix<S> &Q, vector<S> shifts){

  int P = shifts.size();
  int M = Q.dim;
  Q.Unity();

  
  for(int i=0;i<P;i++){
    
    V x, y;
    vector<S> ck(2), v(2);
    
    x = H(con+0,con+0)-shifts[i];
    y = H(con+1,con+0);
    ck[0] = x; ck[1] = y; 
    normalize(ck);
    //cout << fixed << scientific << setprecision(16) << setw(16) << "ck = (" << x << " , " << y << ")" << endl;
    V beta;
    Householder_vector<S>(ck, 0, 1, v, beta);
    //cout << "v = (" << v[0] << " , " << v[1] << ")" << endl;
    Householder_mult_tri<S>(H,v,beta,con+0,con+3,con+0,con+1,0);
    Householder_mult_tri<S>(H,v,beta,con+0,con+3,con+0,con+1,1);
    
    ///Accumulate eigenvector
    Householder_mult<S>(Q,v,beta,0,con+0,con+2,1);
    int sw = 0;

    for(int k=con+0;k<M-2;k++){

      x = H(k+1,k); 
      y = H(k+2,k); 
      if(k+3 > M-1){sw = 1;}

      ck[0] = x; ck[1] = y;
      normalize(ck);
      //cout << "ck = (" << x << " , " << y << ")" << endl;
      Householder_vector<S>(ck, 0, 1, v, beta);
      //cout << "v = (" << v[0] << " , " << v[1] << ")" << endl;
      Householder_mult_tri<S>(H,v, beta,k,k+4-sw,k+1,k+2,0);
      Householder_mult_tri<S>(H,v, beta,k,k+4-sw,k+1,k+2,1);
      ///Accumulate eigenvector
      Householder_mult<S>(Q,v, beta,0,k+1,k+3-sw,1);

    }
    }
  }
  */


  ///Converge on invariant subspace and quit "elegantly"
  template <class S> void Krylov<S>::Abort(int ff, vector<S> &evals,  vector<vector<S> > &evecs){

  QDPIO::cout << "Check for converged eigenvectors before aborting" << endl;
  QDPIO::cout << "Subspace dimension " << ff << " < " << this->M << endl; 

  std::vector<S> tevals(ff-this->con);
  std::vector<std::vector<S> > tevecs(ff-this->con);

  TestConv(ff, tevals, tevecs);
  Matrix<S> AH(ff); 
  AH = this->H.GetSubMtx(0,ff,0,ff);

  this->H.resize(ff);
  this->H = AH;
  
  evals.resize(ff);
  evecs.resize(ff);
  this->K = this->con;
  this->M = ff;
  }

  ///result = gamma5 R_5 input.
  template <class S> void Krylov<S>::G5R(Fermion_t result[2], Fermion_t input[2]){

  int Ls = dop.Ls;
  int x[4];
  int Nspinco=12;
  int cb0, cb1, site, bidx, rb_idx;
  S sgn;
  // PAB
  // Very inefficient
  for(int s=0;s<this->Ls;s++){
  for ( x[3]=0; x[3]<dop.node_latt[3];x[3]++ ) { 
  for ( x[2]=0; x[2]<dop.node_latt[2];x[2]++ ) { 
  for ( x[1]=0; x[1]<dop.node_latt[1];x[1]++ ) { 
  for ( x[0]=0; x[0]<dop.node_latt[0];x[0]++ ) { 

    site = x[0]+x[1]+x[2]+x[3];   
    if(prectype == 1){
    cb0 = ((site+s)&0x1);
    cb1 = ((site+Ls-1-s)&0x1);}
    else{
    cb0 = ((site)&0x1);
    cb1 = ((site)&0x1);}

      for ( int co=0;co<Nspinco;co++ ) { 
      for ( int reim=0;reim<2;reim++ ) { 
	
	bidx = dop.bagel_idx5d(x,Ls-1-s,reim,co,Nspinco,1);
	rb_idx = dop.bagel_idx5d(x,s,reim,co,Nspinco,1);	///TODO Must be an easier way...
	sgn = 1.0;
	if(co > 5) sgn = -1.0;
	
	
if(this->prec == 0){
	S * forward = (S *)input[cb1];
	S * backward = (S *)result[cb0];
	backward[rb_idx] = sgn * forward[bidx]; 
}
else{
  	S * forward = (S *)input[1];
	S * backward = (S *)result[1];
	backward[rb_idx] = sgn * forward[bidx]; 
}

      }}
  }}}}
  }

}

/*
 template <class S> int Krylov<S>::Arnoldi_Factor(int start, int end){

  start--;
  bfm_fermion tmp; 
  bfm_fermion w; 
  bfm_fermion btmp;
  this->init_fermion(tmp);
  this->init_fermion(w);
  this->init_fermion(btmp); 
  double beta;

  if(start == -1){
    start++;
    this->dwf_multiply(this->bq[start],w);
    complex<double> hh = this->innerProduct(this->bq[start],w); 
    this->innerprod++;
    this->axpy(tmp, 0.0, this->bq[start], this->bq[start]);
    this->scale(tmp, hh);
    this->axpy(this->bf[start], -1.0, tmp, w);
    this->H( hh ,start,start);
  }

 
  beta = this->axpy_norm(this->bf[start],0.0,this->bf[start],this->bf[start]);
  double bt = sqrt(beta);

  QDPIO::cout << start << " bt =  " << bt << endl;

  for(int j=start;j<end-1;j++){

    QDPIO::cout << "Factor j " << j+1 << endl;
    #ifndef BLUEGENE
    QDPIO::cerr << "Factor j " << j+1 << endl;
    #endif 
    this->axpby( this->bq[j+1] , (1.0/( bt ) ), this->bf[j], 0.0, this->bf[j] );
    this->H(bt,j+1,j);

    QDPIO::cout << "bt =  " << bt << endl;
    this->dwf_multiply(this->bq[j+1],w);

    multi1d<complex<double> > h(j+2);
    for(int k=0;k<j+2;k++){
      h[k] = this->innerProduct(this->bq[k],w);
    }


    this->axpby(tmp, 0.0, tmp, 0.0, tmp);
    for(int l=0;l<j+2;l++){
      this->axpy(btmp,0.0,this->bq[l],this->bq[l]);
      this->scale(btmp, h[l]);
      this->axpy(tmp,1.0,btmp,tmp);
    }
    this->axpy(this->bf[j+1],-1.0,tmp,w);

    double t = this->axpy_norm(this->bf[j+1],0.0,this->bf[j+1],this->bf[j+1]); 
    double normf = t; 
    bt = sqrt(normf);
    complex<double> tt = 0.0;
    for(int l=0;l<j+2;l++){
      tt = tt + h[l];
    }
    double normh = real(t);

/// Extra orthagonalization step. Better than Gram Schmidt.
/// Can be skipped sometimes but just always do it.
    for(int re=0;re<1;re++){
      multi1d<complex<double> > x(j+2);
      for(int k=0;k<j+2;k++){
	x[k] = this->innerProduct(this->bq[k],this->bf[j+1]);
      }
      
      this->axpby(tmp, 0.0, tmp, 0.0, tmp);
      for(int l=0;l<j+2;l++){
	this->axpy(btmp,0.0,this->bq[l],this->bq[l]);
	this->scale(btmp, x[l]);
	this->axpy(tmp,1.0,btmp,tmp);
      }

      this->axpy(this->bf[j+1] ,-1.0,tmp,this->bf[j+1]);
      for(int l=0;l<j+2;l++){h[l] = h[l] + x[l];}
      
      t = this->axpy_norm(this->bf[j+1],0.0,this->bf[j+1],this->bf[j+1]); 
      normf = t;
      bt = sqrt(normf);
    }
    
    for(int k=0;k<j+2;k++){
      this->H( h[k],k,j+1);
    }

  }
  this->free_fermion(tmp);
  this->free_fermion(w);
  this->free_fermion(btmp); 
  return end;
}*/


  template <class S> int Krylov<S>::Lanczos_Factor(int start, int end){

  S beta;  
  complex<S> alpha;

  QDPIO::cout<<"Lanczos_Factor start/end " <<start <<"/"<<end<<endl;
  ///Starting from scratch, bq[0] contains a random vector and |bq[0]| = 1
  if(start == 0){
    start++;
    Chebyshev(this->bq[0],this->bf[0]);						//bf = A bq[0]
    alpha = this->innerProduct(this->bq[0],this->bf[0]);this->innerprod++;	//alpha =  bq[0]^dag A bq[0]
    this->axpy(this->bf[0] ,-1.0*real(alpha), this->bq[0], this->bf[0]);	//bf =  A bq[0] - alpha bq[0]
    this->H( real(alpha),0,0);							//bq[0] . bf =  bq[0]^dag A bq[0] - alpha |bq[0]|^2 = 0
  }
  int re = 0;
  start--;

  beta = this->axpy_norm(this->bf[0],0.0,this->bf[0],this->bf[0]);		//|bf|^2
  S sqbt = sqrt(beta);

  if( this->cont){beta = 0;sqbt = 0;}	
  for(int j=start+1;j<end;j++){
    QDPIO::cerr << "Factor j " << j+1 << endl;
    if(this->cont){
	this->axpy(this->bq[j] , 0.0 , this->bf[0], this->bf[0]);this->cont = false;
    }
    else{
      this->axpby(this->bq[j] , (1.0/sqbt) , this->bf[0], 0.0, this->bf[0]);	//bq = bf/|bf|
      this->H( sqbt,j,j-1); this->H( sqbt,j-1,j);
    }

    Chebyshev(this->bq[j],this->bf[0]);									//bf = A bq[j]
    this->axpy(this->bf[0], -1.0*sqbt, this->bq[j-1], this->bf[0]);					//bf = A bq[j] - beta bq[j-1]
    alpha = this->innerProduct(this->bq[j],this->bf[0]); this->innerprod++;				//alpha = bq[j]^dag A bq[j]
    S fnorm = this->axpy_norm(this->bf[0] , -1.0*real(alpha), this->bq[j], this->bf[0]);		//bf = A bq[j] - beta bq[j-1] - alpha bq[j]

    re = 0;
    S bck = sqrt( real( conj(alpha)*alpha ) + beta );
    beta = fnorm;
    sqbt = sqrt(beta);

    ///Iterative refinement of orthogonality V = [ bq[0]  bq[1]  ...  bq[M] ]
    while( re == this->ref || (sqbt < this->rho * bck && re < 5) ){

      //bex = V^dag bf
      vector<complex<S> > bex(j+1);
      for(int k=0;k<j+1;k++){bex[k] = this->innerProduct(this->bq[k],this->bf[0]); this->innerprod++;}

      this->zero_fermion(this->tmp2);
      //tmp2 = V s
      for(int l=0;l<j+1;l++){
        S nrm = norm2(this->bq[l]);
        this->axpy(this->tmp1,0.0,this->bq[l],this->bq[l]); this->scale(this->tmp1,bex[l]); 	//tmp1 = V[j] bex[j]
        this->axpy(this->tmp2,1.0,this->tmp2,this->tmp1);					//tmp2 += V[j] bex[j]
      }

      //bf = bf - V V^dag bf.   Subtracting off any component in span { V[j] } 
      S btc = this->axpy_norm(this->bf[0],-1.0,this->tmp2,this->bf[0]);
      alpha = alpha + bex[j];	      sqbt = sqrt(real(btc));	      
      S nmbex = 0;for(int k=0;k<j+1;k++){nmbex = nmbex + real( conj(bex[k])*bex[k]  );}
      bck = sqrt( nmbex );
      re++;
    }
    this->H( real(alpha) ,j,j);
    if(re > 1) QDPIO::cout << "orthagonality refined " << re << " times" << endl;
  }
  return end;
}

  /**Print out the true residuals**/
  template <class S> void Krylov<S>::True_Residual( void ){
      
    std::vector<S> mevals(this->M);
    std::vector<std::vector<S> > mevecs(this->M);
    Wilkinson<S>(this->H, mevals, mevecs, this->convQR);
    EigenSort(mevals,mevecs,this->comp);
  
      bfm_fermion bS; this->init_fermion(bS); 
      bfm_fermion bv; this->init_fermion(bv);
      bfm_fermion bvv; this->init_fermion(bvv);
      
    for(int i=this->M-1; i>(this->M-1-this->get) && i>=0; i--){
     
      this->axpby(bS,0.0,bS,0.0,bS);

      double bdiff;
      S eval_i;
      if(this->kr == lan){

	for(int j=0;j<this->bq.size();j++){
  		this->axpy(bS,real(mevecs[i][j]), this->bq[j], bS);
	}
	
  		this->herm_mult(bS,bv);
  		eval_i = this->innerProduct_real(bS,bv);	///Rayleigh quotient
  		double bS_norm = sqrt( this->axpy_norm(bS,0,bS,bS) );
  		eval_i = eval_i/bS_norm;
  		this->axpby(bS,0,bS,1.0/bS_norm,bS);
  		double bv_norm = sign( real(eval_i) ) * sqrt( this->axpy_norm(bv,0,bv,bv) );
  		this->axpby(bv,0,bv,1.0/bv_norm,bv);

  		bdiff = this->axpy_norm(bvv, -1.0 , bS, bv);
	}


      bdiff = sqrt(bdiff);
      QDPIO::cout <<"true residual " << this->M-1-i << " = " <<  bdiff << " of "
		  << "  :  (" << real(eval_i) << "," << imag(eval_i) << ")" << endl;
      
    }
      this->free_fermion(bS); 
      this->free_fermion(bv);
      this->free_fermion(bvv);
   
  }
  
  

  template <class S> void Krylov<S>::TestConv(int SS, vector<S> &tevals, vector<vector<S> > &tevecs){
  
  QDPIO::cout << "Converged " << this->con << " so far." << endl;
  ///Active Factorization
  Matrix<S> AH(SS-this->con);
  AH = this->H.GetSubMtx(this->con, SS,this->con, SS);
  
  //if(this->kr == arn){ QReigensystem<S>(AH, tevals, tevecs, this->convQR);}
  if(this->kr == lan){ Wilkinson<S>(AH, tevals, tevecs, this->convQR);}
  EigenSort(tevals,tevecs,this->comp);
  
  Real resid_nrm;
  //if(this->kr == arn) resid_nrm =  this->axpy_norm(bf[SS-1],0.0,bf[SS-1],bf[SS-1]) ;
  if(this->kr == lan) resid_nrm =  this->axpy_norm(bf[0],0.0,bf[0],bf[0]) ;
  int oldcon = this->con;


  for(int i=SS-this->con-1; i>=SS-this->K && i >= 0  ;i--){
   
    double diff = 0;
    diff = abs(tevecs[i][this->M-1-this->con])*toDouble(resid_nrm); 
    
    QDPIO::cout << "residual estimate " << SS-1-i << " " << diff << " of (" << real(tevals[i]) 
		<< " , " << imag(tevals[i]) << ")" << endl;
    
    if(  (diff < this->conv && (i==SS-this->con-1) ) ){
      Matrix<S> Q(this->M);
      
      

      if(this->con < this->get){
	this->converged.push_back(tevals[i]);
	if(this->lock){
	  bool herm = true; 
	  //if(this->kr == arn){ herm = false;}  
	  Deflate<S>::Lock(this->H, Q, tevals[i], this->con, this->convQR, SS, herm);
	  /*if(this->kr == arn){
	    this->times(this->bq,Q,this->bq.size());
	    this->scale(this->bf[this->M-1] , Q(this->M-1,this->M-1) );
	  }*/
	  if(this->kr == lan){
	    this->times_real(this->bq,Q,this->bq.size());
	    this->axpby(this->bf[0] ,  real( Q(this->M-1,this->M-1) ), this->bf[0], 0.0, this->bf[0] );
	  }

	}
	this->con++;
	QDPIO::cerr << " converged on eval " << this->con << " of " << this->get << endl;
      }
    }
  }

  QDPIO::cout << "Got " << this->con << " so far " << endl;	
  QDPIO::cerr << "Got " << this->con << " so far " << endl;	
  
}


  template <class S> void Krylov<S>::ImplicitRestart(	int TM, vector<S> &evals,  vector<vector<S> > &evecs){

  /// Sort by smallest imaginary part or whatever
  EigenSort(evals,evecs,this->comp);
  ///Assign shifts
  if(this->K - this->con < 4) this->P = (this->M - this->K-1); //one
  //if(this->K - this->con == 1) this->P = (this->M - this->K - ( (this->M -this->K)/2 ) ); //two
  //if(this->K - this->con < (this->M -this->K)/2 ) this->P = (this->M - this->K - ( (this->M -this->K)/2 ) ); //three Best?
  //if(this->K - this->con == 1) this->P = (this->M - this->K-1); //one
  
  std::vector<S> shifts(this->P + this->shift_extra.size() );
  for(int k=0;k<this->P;k++){shifts[k] = evals[k]; }
    
  /// Shift to form a new H and q
  Matrix<S> Q(TM);Q.Unity();
  //for(int k=0;k<shifts.size();k++){cout << "shift " << k << " " << shifts[k] << endl; }
  Shift(Q,shifts);
  
  int ff = this->K;
  /// Shifted H defines a new K step Arnoldi factorization
  S  beta = this->H(ff,ff-1); 
  S  sig = Q(TM-1,ff-1);
  QDPIO::cout << "beta = " << real(beta) << " sig = " << real(sig) << endl;
  /*if(this->kr == arn) {
    /// q -> q Q TODO
    this->times(this->bq,Q, TM);
    this->axpy(this->tmp1,0.0,this->bq[this->K],this->bq[this->K]);
    this->scale(this->tmp1, beta);
    this->axpy(this->tmp2,0.0,this->bf[TM-1],this->bf[TM-1]);
    this->scale(this->tmp2, sig);
    this->axpy(this->bf[this->K-1],1.0,this->tmp1,this->tmp2);
    /// Do the rest of the factorization
    ff = Arnoldi_Factor(ff,this->M);
  }*/
  if(this->kr == lan) {
    QDPIO::cout << "TM = " << TM << endl;
    /// q -> q Q
    QDPIO::cout << this->norm2(this->bq[0]) << " -- before" << endl;

    this->times_real(this->bq,Q, TM);
    QDPIO::cout << this->norm2(this->bq[0]) << " -- after " << ff << endl;
    this->axpby(bf[0],real(beta),this->bq[ff],real(sig),this->bf[0]);
    /// Do the rest of the factorization
    ff = Lanczos_Factor(ff,this->M);
  }

  if(ff < this->M){Abort(ff,evals,evecs);}

  }


   ///Check
   template <class S> void Krylov<S>::Check(){
   /*vector<multi1d<LatticeFermion> > goodvec(0);*/
   vector<S> goodval(this->K);
   EigenSort(this->evals,this->evecs,this->comp);
   multi1d<bfm_fermion> con_vecs(this->K); for(int i=0;i<this->K;i++){this->init_fermion(con_vecs[i]);}
     bfm_fermion bS; this->init_fermion(bS); 
     bfm_fermion bv; this->init_fermion(bv);  
     bfm_fermion bvv; this->init_fermion(bvv);
   
   for(int i=this->M-1;i>this->M - this->K-1 && i>=0;i--){

     int ii = this->M-1-i;
     this->axpby(bS,0.0,bS,0.0,bS);
     double bdiff ;
     if(this->kr == lan){
       for(int j=0;j<this->bq.size();j++){this->axpy(bS,real(this->evecs[i][j]), this->bq[j], bS);}
       this->herm_mult(bS,bv);
       goodval[ii] = this->innerProduct_real(bS,bv);	///Rayleigh quotient
        double bS_norm = sqrt( this->axpy_norm(bS,0,bS,bS) );
  	goodval[ii] = goodval[ii]/bS_norm;
  	this->axpby(bS,0,bS,1.0/bS_norm,bS);
  	double bv_norm = sign( real(goodval[ii]) ) * sqrt( this->axpy_norm(bv,0,bv,bv) );
  	this->axpby(bv,0,bv,1.0/bv_norm,bv);	
        bdiff = this->axpy_norm(bvv, -1.0 , bS, bv);
        this->axpy(con_vecs[ii],0.0,bS,bS);
     }
     /*if(this->kr == arn){
       for(int j=0;j<this->bq.size();j++){
	 this->axpy(this->tmp1, 0.0, this->bq[j], this->bq[j]);
	 this->scale(this->tmp1,evecs[i][j]);
	 bdiff = this->axpy_norm(bS,1.0, this->tmp1, bS);
       }
       this->dwf_multiply(bS,bv);
       this->axpy(this->tmp1, 0.0, bS, bS);
       this->scale(this->tmp1,evals[i]);
       bdiff = this->axpy_norm(bvv, -1.0 , this->tmp1, bv);
     }*/
     bdiff = sqrt(bdiff);

     string cv= " not converged";  
     if(bdiff < 10*this->conv ){ 
       cv = " converged";
     }else{
       QDPIO::cout << "WARNING LARGE RESIDUAL = " << bdiff 
		   << " on eval " << ii << endl;
     }
     QDPIO::cout <<"diff = " <<  bdiff << " < " << 10*this->conv 
		 << "  :  (" << real(goodval[ii]) << "," << imag(goodval[ii]) 
		 << ") eval # " << ii << cv << endl;




   }
     this->free_fermion(bS);
     this->free_fermion(bv);
     this->free_fermion(bvv);

   this->evals.resize(this->K);
   this->bl.resize(this->K);

   for(int i=0;i<this->K;i++){
	this->axpy(this->bq[i],0.0,con_vecs[i],con_vecs[i]); 
	this->evals[i] = goodval[i]; 
	this->bl[i] = goodval[i]; 
	this->free_fermion(con_vecs[i]);
   }

   QDPIO::cout << "total number of vector inner products " << this->innerprod << endl; 
   QDPIO::cout << "total number of matrix vector products " << this->mvprod << endl;
 }

 template <class S> void Krylov<S>::multiply(bfm_fermion input, bfm_fermion &result){
	EigenType tmp = this->D;
	if(this->D == G5D){ this->D = DDAGD;}
	herm_mult(input, result);
	this->D = tmp;
 }
 template <class S> void Krylov<S>::shift_multiply(bfm_fermion input, bfm_fermion &result){
	bfm_fermion v0; this->init_fermion(v0);
	EigenType tmp = this->D;
	
	herm_mult(input, v0);
	this->axpy(v0, -this->ch_shift, input, v0);
	herm_mult(v0, result);
	this->axpy(result, -this->ch_shift, v0, result);
	
	this->free_fermion(v0);
	this->D = tmp;
 }
 ///q(Q)= 2*( Q^2/s^2 ) - ( 1+r ) I
 template <class S> void Krylov<S>::qpoly(double beta, double alpha, bfm_fermion input, bfm_fermion &result){
	
	if(this->sh){
	  this->shift_multiply(input,result);
	}
	else{
	  this->multiply(input,result);
	}
	alpha = alpha*alpha;
	beta = beta*beta;
	double amb = (alpha - beta);
	double apb = (alpha + beta);
	this->axpby(result, 2.0/amb, result, -apb/amb, input);
 }
  template <class S> double Krylov<S>::qpoly(double alpha, double beta, double x){
	return (2.0*x*x - (alpha + beta)) / (beta - alpha);
 }

 /// n order Chebyshev
 template <class S> void Krylov<S>::Chebyshev(bfm_fermion input, bfm_fermion &v1){

  //if(this->kr == arn) this->dwf_multiply(input,v1); 
  if(Tn == 0){
	if(this->kr == lan) this->herm_mult(input,v1); 
	return;
  }
  bfm_fermion v0; this->init_fermion(v0); 
  this->axpy(v0,0.0,input, input);

	this->qpoly(ch_rad,ch_off,v0,v1);

	if(Tn == 1){this->free_fermion(v0); return;}
	bfm_fermion tmp; this->init_fermion(tmp);

	for(int i=1;i<Tn;i++){
		this->qpoly(ch_rad,ch_off,v1,tmp);
		this->axpby(tmp,2.0,tmp,-1.0,v0);
		this->axpy(v0,0.0,v1,v1);
		this->axpy(v1,0.0,tmp,tmp);
	}

	this->free_fermion(tmp);
	this->free_fermion(v0);
 }
 /// n order Chebyshev 
 template <class S> double Krylov<S>::Chebyshev(double x){
 
	double v0 = 1;
	if(Tn == 0){return x;}
	double v1 = this->qpoly(ch_off, ch_rad, x);
	double xbar = v1;
	if(Tn == 1){return v1;}
	double tmp;
  
	for(int i=1;i<Tn;i++){
		tmp = 2.0*xbar*v1-1.0*v0;
		v0=v1;
		v1=tmp;
	}

	return v1;
}


 template <class S> complex<double> Krylov<S>::innerProduct(Fermion_t x[2], Fermion_t y[2] ){
   complex<double> dot;
#pragma omp parallel 
  {
    omp_set_num_threads(dop.nthread);
#pragma omp for 
    for(int i=0;i<dop.nthread;i++) {
     complex<double> mydot = 0;
     for(int cb = this->prec; cb < 2; cb++){
     mydot += dop.inner(x[cb],y[cb]);
     }
     dot=mydot;
   }}
   return dot;
 }
 template <class S> double Krylov<S>::innerProduct_real(Fermion_t x[2], Fermion_t y[2] ){
   double nrm;
#pragma omp parallel 
  {
    omp_set_num_threads(dop.nthread);
#pragma omp for 
    for(int i=0;i<dop.nthread;i++) {
     double mynrm = 0;
     for(int cb = this->prec; cb < 2; cb++){
     mynrm+=dop.inner_real(x[cb],y[cb]) ;
     }
     nrm = mynrm;
   }}
   return nrm;
 }
  template <class S> void Krylov<S>::gamma5(Fermion_t x[2]){ 
#pragma omp parallel 
  {
    omp_set_num_threads(dop.nthread);
#pragma omp for 
    for(int i=0;i<dop.nthread;i++) {
    for(int cb=this->prec;cb<2;cb++){
    dop.chiralproj_axpby(x[cb], x[cb], -2.0 , -1,1);
    }
  }}
}
 template <class S> double Krylov<S>::norm2(Fermion_t x[2]){
   double nrm;
#pragma omp parallel 
  {
    omp_set_num_threads(dop.nthread);
#pragma omp for 
    for(int i=0;i<dop.nthread;i++) {
     double mynrm = 0;
     for(int cb=this->prec;cb<2;cb++){
        mynrm += dop.norm(x[cb]);
     }
     nrm=mynrm;
   }}
   return nrm;
 }
 template <class S> double Krylov<S>::norm(Fermion_t x[2]){
   return sqrt(norm2(x));
 }

template <class S> void Krylov<S>::scale(Fermion_t x[2], complex<double> s){ 
#pragma omp parallel 
  {
    omp_set_num_threads(dop.nthread);
#pragma omp for 
    for(int i=0;i<dop.nthread;i++) {
    for(int cb=this->prec;cb<2;cb++){
    dop.scale(x[cb],real(s),imag(s));
    }
  }}
}
template <class S> void Krylov<S>::scale(Fermion_t x[2], S s){ 
#pragma omp parallel 
  {
    omp_set_num_threads(dop.nthread);
#pragma omp for 
    for(int i=0;i<dop.nthread;i++) {
    for(int cb=this->prec;cb<2;cb++){
    dop.scale(x[cb],s);
    }
  }}
}

template <class S> void Krylov<S>::equate(Fermion_t x[2], Fermion_t y[2] ){
#pragma omp parallel 
  {
    omp_set_num_threads(dop.nthread);
#pragma omp for 
    for(int i=0;i<dop.nthread;i++) {
    for(int cb=this->prec;cb<2;cb++){ 
      dop.equate(x[cb],y[cb]);  
    }
  }}
}
template <class S> void Krylov<S>::init_fermion(Fermion_t x[2]){
#pragma omp parallel 
  {
    omp_set_num_threads(dop.nthread);
#pragma omp for 
    for(int i=0;i<dop.nthread;i++) {
  for(int cb=this->prec;cb<2;cb++){
    x[cb]  =  dop.threadedAllocFermion();
  } 
  }
}
}
template <class S> void Krylov<S>::zero_fermion(Fermion_t x[2]){

#pragma omp parallel 
  {
    omp_set_num_threads(dop.nthread);
#pragma omp for 
    for(int i=0;i<dop.nthread;i++) {
    for(int cb=this->prec;cb<2;cb++){
      dop.set_zero(x[cb]);
    } 
  }}
}
template <class S> void Krylov<S>::free_fermion(Fermion_t x[2]){
#pragma omp parallel 
  {
    omp_set_num_threads(dop.nthread);
#pragma omp for 
    for(int i=0;i<dop.nthread;i++) {
  for(int cb=this->prec;cb<2;cb++){
    dop.threadedFreeFermion(x[cb]);
  } 
}}

}
template <class S> void Krylov<S>::qdp_to_bfm(LatticeFermion in, Fermion_t out[2]){
  for(int cb=this->prec;cb<2;cb++){
    dop.importFermion(in,out[cb],cb);
  }
}
template <class S> void Krylov<S>::qdp_to_bfm(multi1d<LatticeFermion> &in, Fermion_t out[2]){
  for(int cb=this->prec;cb<2;cb++){
    dop.importFermion(in,out[cb],cb);
  }
}
template <class S> void Krylov<S>::bfm_to_qdp(Fermion_t in[2], LatticeFermion &out){
  for(int cb=this->prec;cb<2;cb++){
    dop.exportFermion(out,in[cb],cb);
  }
}
template <class S> void Krylov<S>::bfm_to_qdp(Fermion_t in[2], multi1d<LatticeFermion> &out){
  for(int cb=this->prec;cb<2;cb++){
    dop.exportFermion(out,in[cb],cb);
  }
}
template <class S> double Krylov<S>::axpby_norm(Fermion_t r[2], double a, Fermion_t x[2], double b, Fermion_t y[2])
{
  double sum = 0;
#pragma omp parallel 
  {
    omp_set_num_threads(dop.nthread);
#pragma omp for 
    for(int i=0;i<dop.nthread;i++) {
    double mysum = 0;
    for(int cb=this->prec;cb<2;cb++){
    mysum += dop.axpby_norm( r[cb], x[cb], y[cb], a, b);
    }
    sum=mysum;
  }}
  return sum;
}
template <class S> void Krylov<S>::axpby(Fermion_t r[2], double a, Fermion_t x[2],double b, Fermion_t y[2])
{
#pragma omp parallel 
  {
    omp_set_num_threads(dop.nthread);
#pragma omp for 
    for(int i=0;i<dop.nthread;i++) {
    for(int cb=this->prec;cb<2;cb++){
      dop.axpby( r[cb], x[cb], y[cb], a, b);
    }
  }}
}
template <class S> double Krylov<S>::axpy_norm(Fermion_t r[2], double a, Fermion_t x[2], Fermion_t y[2]){
  double sum;
#pragma omp parallel 
  {
    omp_set_num_threads(dop.nthread);
#pragma omp for 
    for(int i=0;i<dop.nthread;i++) {
    double mysum = 0;
    for(int cb=this->prec;cb<2;cb++){
    mysum += dop.axpy_norm( r[cb], x[cb], y[cb], a);
    }
    sum=mysum;
  }}
  return sum;
}
template <class S> void Krylov<S>::axpy(Fermion_t r[2], double a, Fermion_t x[2], Fermion_t y[2]){
#pragma omp parallel 
  {
    omp_set_num_threads(dop.nthread);
#pragma omp for 
    for(int i=0;i<dop.nthread;i++) {
    for(int cb=this->prec;cb<2;cb++){
      dop.axpy( r[cb], x[cb], y[cb], a);
    }
  }}
}
/// q -> q Q
template <class S> void Krylov<S>::times(multi1d<bfm_fermion> &q, Matrix<S> &Q, int N){

  multi1d<bfm_fermion> SS( N ); 
  for(int i=0;i<N ;i++){
    this->init_fermion(SS[i]);
  }
  bfm_fermion tmp; 
  this->init_fermion(tmp); 
  

  for(int j=0;j<N ;j++){
    this->axpby(SS[j],0.0,SS[j],0.0,SS[j]);
    for(int k=0;k<N ;k++){
      this->axpy(tmp,0.0,q[k],q[k]);
      this->scale(tmp,Q(k,j));
      this->axpy(SS[j],1.0,SS[j],tmp);
    }
  }
  
  for(int j=0;j<N ;j++){
    axpy(q[j],0.0,SS[j],SS[j]);
  }
  
  this->free_fermion(tmp);
  for(int j=0;j<N ;j++){
    this->free_fermion(SS[j]);
  }
}
/// q -> q Q
template <class S> void Krylov<S>::times_real(multi1d<bfm_fermion> &q, Matrix<S> &Q, int N){

  multi1d<bfm_fermion> SS( N ); 
	  
  for(int j=0;j<N ;j++){
    this->init_fermion(SS[j]); this->zero_fermion(SS[j]);
    for(int k=0;k<N ;k++){
      if(real( Q(k,j) ) != 0.0){
	this->axpy(SS[j], real( Q(k,j) ),q[k],SS[j]);
      }
    }
  }

	  
  for(int j=0;j<N ;j++){
    this->axpy(q[j],0.0,SS[j],SS[j]);     this->free_fermion(SS[j]);
  }
  
}
/// Gram Schmidting.
template <class S> void Krylov<S>::Gram(Fermion_t r[2], multi1d<bfm_fermion> &q, int N){

  bfm_fermion tmp; 
  this->init_fermion(tmp); 

  {
      for(int i=0;i<N;i++){
      double norm = this->axpy_norm(tmp, 0.0, q[i], q[i]);
      complex<double> sub = ( this->innerProduct(q[i],r)/norm );
      this->scale(tmp,sub);
      this->axpy(r , -1.0 , tmp, r);
      }
    
    double nrm = this->norm(r);
    this->axpby(r,  1.0/nrm,r, 0.0,r );
  }
  this->free_fermion(tmp);
}

  template <class S> void Krylov<S>::ObliqueProj(Fermion_t x[2], Fermion_t b[2]){

QDPIO::cout << "sol norm 1 cpt = " << this->norm(x) << endl;
QDPIO::cout << "source norm 1 cpt = " << this->norm(b) << endl;

	std::vector<complex<S> > Wb(this->K);
	for(int k=0;k<this->K; k++){
		Wb[k] = this->innerProduct(this->bq[k],b);
		
	}
	multi1d<bfm_fermion> AW(this->K);
	for(int k=0;k<this->K; k++){
		this->init_fermion(AW[k]);
		this->herm_mult(this->bq[k],AW[k]);
	}
	
	Matrix<complex<S> > WAW(this->K);
	for(int j=0;j<this->K; j++){
	for(int k=0;k<this->K; k++){
		WAW(this->innerProduct(this->bq[j],AW[k]),j,k);
	}}

	WAW.Out();

	std::vector<complex<S> > WAWWb(this->K);
	print(Wb);
	CG_Matrix(WAW,Wb,WAWWb);
	print(WAWWb);
	QDPIO::cout << "norm x " << this->norm(x) << endl;
	this->axpby(x,0.0,x,0.0,x);
	bfm_fermion tmpt;
	this->init_fermion(tmpt);
	for(int k=0;k<this->K; k++){
		this->axpy(tmpt,0,this->bq[k],this->bq[k]);
		this->scale(tmpt,WAWWb[k]);
		this->axpy(x,1.0,x,tmpt);
	QDPIO::cout << "norm x " << this->norm(x) << endl;
	}
	this->free_fermion(tmpt);
	QDPIO::cout << "norm x " << this->norm(x) << endl;
	for(int k=0;k<this->K; k++){
		this->free_fermion(AW[k]);
	}

  }

#endif


